# coding utf-8

# Built-in
import itertools as itt
import warnings

# Common
import numpy as np
# import scipy.integrate as scpinteg
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib as mpl
import matplotlib.transforms as transforms
import matplotlib.lines as mlines
# from mpl_toolkits.axes_grid1 import make_axes_locatable

# tofu
from tofu.version import __version__
from .. import utils as utils
from . import _def as _def
from . import _generic_check


__all__ = [
    'plot_TimeTraceColl',
    'plot_axvline',
]


# __author_email__ = 'didier.vezinet@cea.fr'
__github = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at {}'.format(
    __version__, __github,
)
_nchMax, _ntMax, _nfMax, _nlbdMax = 4, 3, 3, 3
_fontsize = 8
_labelpad = 0
_lls = ['-', '--', '-.', ':']
_lct = [plt.cm.tab20.colors[ii] for ii in [0, 2, 4, 1, 3, 5]]
_lcch = [plt.cm.tab20.colors[ii] for ii in [6, 8, 10, 7, 9, 11]]
_lclbd = [plt.cm.tab20.colors[ii] for ii in [12, 16, 18, 13, 17, 19]]
_lcm = _lclbd
_cbck = (0.8, 0.8, 0.8)
_dmarker = {'ax': 'o', 'x': 'x'}


_OVERHEAD = True
_CROSS = True
_DRAW = True
_CONNECT = True
_AXGRID = False
_LIB = 'mpl'
_BCKCOLOR = 'w'


#############################################
#############################################
#       TimeTraceCollection plots
#############################################
#############################################


def _check_proj(proj=None, allowed=None):
    """ Return validated proj as a list of keys """

    # Check allowed
    c0 = (
        isinstance(allowed, list)
        and all([isinstance(ss, str) for ss in allowed])
    )
    if not c0:
        msg = (
            "Arg allowed must be list of str!\n"
            + "\t- provided: {}".format(allowed)
        )
        raise Exception(msg)

    # Check proj
    if proj is None:
        proj = 'all'
    if proj == 'all':
        proj = allowed
    if isinstance(proj, str):
        proj = [proj]
    c0 = (
        isinstance(proj, list)
        and all([isinstance(ss, str) and ss in allowed for ss in proj])
    )
    if not c0:
        msg = (
            "Arg proj must be a list of str from {}\n".format(allowed)
            + "\t- provided: {}".format(proj)
        )
        raise Exception(msg)

    return proj


#############################################
#############################################
#       TimeTraceCollection plots
#############################################
#############################################


def _get_fig_dax_mpl(dcases=None, axgrid=None,
                     overhead=False, novert=1, noverch=1,
                     ntmax=None, nchmax=None,
                     sharex_t=None, sharey_t=None,
                     sharex_ch=None, sharey_ch=None,
                     cross=False, share_cross=None, cross_unique=None,
                     bckcolor=None, wintit=None, tit=None):

    # -----------------
    # Check format input
    # -----------------

    assert isinstance(nax, int)
    lc = [axgrid is None, axgrid is True, isinstance(axgrid, tuple)]
    assert any(lc)

    if bckcolor is None:
        bckcolor = _BCKCOLOR
    if wintit is None:
        wintit = _WINTIT

    if axgrid is None:
        axgrid = _AXGRID
    if axgrid is None:
        axgrid = False
    if axgrid is not False:
        assert dim == 1
    if dmargin is None:
        dmargin = _def.dmargin1D

    # -----------------
    # Check all cases
    # -----------------

    if axgrid is not False:
        # Only time traces
        assert all([vv['dim'] == 1 for vv in dcases.values()])
        ncases = len(dcases)
    elif any([vv.get('isspectral', False) for vv in dcases.values()]):
        # Spectral case => only one case !
        assert len(dcases) == i
        isspectral = True
    else:
        # All time traces to overhead, not counted in ncases
        ncases = 0
        for kk, vv in dcases.items():
            if vv['dim'] > 1:
                if vv['is2d']:
                    dcases[kk]['naxch'] = nchmax
                else:
                    dcases[kk]['naxch'] = 1
                ncases += 1
        isspectral = False

    # Options
    # (A) time traces vignettes
    # (B) time traces with or w/o cross + overhead
    # (C) profiles1d / cam1d with or w/o cross + overhead
    # (D) cam2d with or w/o cross + overhead, nt = 1, 2
    # (E) cam1dspectral with or w/o cross + overhead, nch = 1, 2
    # (F) profile2d with or w/o cross + overhead, nch = 1, 2

    # -----------------
    # Make figure
    # -----------------

    fs = utils.get_figuresize(fs)
    fig = plt.figure(facecolor=bckcolor, figsize=fs)
    if wintit is not False:
        fig.canvas.manager.set_window_title(wintit)

    # -----------------
    # Check all cases
    # -----------------

    dax = {'lkey': 0, 'dict': {}}
    if axgrid is False:
        if isspectral:
            naxvgrid = nchMax + 1 + overhead
        else:
            naxvgrid = ncases + overhead
        naxhgrid = 2*2 + cross
        gridax = gridspec.GridSpec(2*naxvgrid, naxhgrid, **dmargin)

        if cD and ntMax == 2:
            naxh = naxhgrid + 1

        # Create overead t and ch
        if overhead:
            for ii in range(novert):
                key = 'over-t{}'.format(ii)
                i0, i1 = ii*novert, (ii+1)*novert
                dax['dict'][key] = fig.add_subplot(gridax[i0:i1, :2])
            for ii in range(noverch):
                key = 'over-ch{}'.format(ii)
                i0, i1 = ii*noverch, (ii+1)*noverch
                dax['dict'][key] = fig.add_subplot(gridax[i0:i1, 2:4])

            # Add hor
            if cross:
                dax['dict']['hor'] = fig.add_subplot(gridax[:2, 4:])

        # Create cross
        i0 = 2*overhead
        if cross:
            if cross_unique:
                dax['dict']['cross'] = fig.add_subplot(gridax[i0:, 4:])
            else:
                for ii in range(ncases):
                    key = 'cross{}'.format(ii)
                    ii0 = i0+ii*2
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 4:])

        # Create time and channel axes
        for ii in range(ncases):
            key = 't{}'.format(ii)
            ii0 = i0+ii*2
            dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, :2])

        dax['ch'] = []
        for ii in range(ncases):
            if dcases[lcases[ii]].get('is2D', False):
                ii0 = i0+ii*2
                for jj in range(nchmax):
                    key = 'ch{}-{}'.format(ii, jj)
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 2+jj])
            else:
                for ii in range(ncases):
                    key = 'ch{}'.format(ii)
                    ii0 = i0+ii*2
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 2:4])

    else:
        if axgrid is True:
            nax = int(np.ceil(np.sqrt(ncases)))
            naxvgrid = nax
            naxhgrid = int(np.ceil(ncases / nax))
        else:
            naxvgrid, naxhgrid = axgrid
        assert naxvgrid*naxhgrid >= ncases
        axgrid = gridspec.GridSpec(naxvgrid, naxhgrid+cross, **dmargin)
        for ii in range(ncase):
            i0 = ii % naxvgrid
            i1 = ii - i0*naxhgrid
            key = 't{}-{}'.format(i0, i1)
            dax['dict'][key] = fig.add_subplot(gridax[i0, i1])

        if cross:
            dax['dict']['cross'] = fig.add_subplot(gridax[:, -1])

    dax['lkey'] = sorted(list(dax['dict'].keys()))
    dax['fig'] = fig
    dax['can'] = fig.canvas
    return dax


def plot_DataColl(coll, overhead=None,
                  color=None, ls=None, marker=None, ax=None,
                  cross=None, share_cross=None, cross_unique=None,
                  axgrid=None, dmargin=None, legend=None,
                  fs=None, draw=None, connect=None, lib=None):

    # --------------------
    # Check / format input
    # --------------------

    if overhead is None:
        overhead = _OVERHEAD
    if cross is None:
        if 'Plasma' in coll.__class__.__name__:
            cross = _CROSS
        else:
            cross = False
    if share_cross is None and cross:
        share_cross = False
    if draw is None:
        draw = _DRAW
    if connect is None:
        connect = _CONNECT
    if lib is None:
        lib = _LIB

    assert lib == 'mpl', 'Only matplotlib available so far !'

    # --------------------
    # Get keys of data to plot
    # --------------------

    if len(coll.dgroup) == 1:
        # TimeTraces
        dcases = None
    else:
        pass

    # --------------------
    # Get graphics dict of keys
    # --------------------

    # Case with time traces only
    daxg, lparam = {}, coll.lparam
    laxt = [('ax', ax), ('color', color), ('ls', ls), ('marker', marker)]
    for ss, vv in laxt:
        if vv is None:
            daxg[ss] = None
            continue

        if vv in lparam:

            # get keys with matching vv
            lp = coll.get_param(vv, key=lk, returnas=str)
            dv = {pp: [kk for kk in lk if self._ddata['dict'][kk][pp] == pp]
                  for pp in set(lp)}

            # Set new param ss
            if ss not in lparam:
                coll.add_param(ss, value=None)      # TBF

        daxg[ss] = dv

    # Case with any type of data => only valid for time traces (True ?)

    # Get number of axes

    # --------------------
    # Prepare figure / axes
    # --------------------

    # Get array of axes positions as a dict
    dim = len(coll.lgroup)
    config = None
    spectral = coll.isspectral

    if lib == 'mpl':
        daxg = _get_fig_daxg_mpl(dcases=dcases, axgrid=axgrid,
                                 cross=cross, overhead=overhead,
                                 ntmax=ntmax, nchmax=nchmax)

    # --------------------
    # Populate axes with static
    # --------------------

    # --------------------
    # Populate axes with dynamic (dobj)
    # --------------------
    dobj = {}

    # --------------------
    # Interactivity
    # --------------------
    collplot = None

    return collplot


# #############################################################################
# #############################################################################
#                       plot_as_matrix
# #############################################################################


def _plot_as_matrix_check(
    coll=None,
    key=None,
    ind=None,
    cmap=None,
    vmin=None,
    vmax=None,
    aspect=None,
    dcolorbar=None,
    dleg=None,
    data=None,
):

    # key
    lk = [kk for kk, vv in coll.ddata.items() if vv['data'].ndim == 2]
    key = _generic_check._check_var(
        key, 'key', default=None, types=str, allowed=lk,
    )

    # ind
    ind = _generic_check._check_var(
        ind, 'ind', default=[0, 0], types=(list, tuple, np.ndarray),
    )
    c0 = (
        len(ind) == 2
        and all([
            np.isscalar(ii) and isinstance(ii, (int, np.integer))
            for ii in ind
        ])
    )
    if not c0:
        msg = (
            "Arg ind must be an iterable of 2 integer indices!\n"
            f"Provided: {ind}"
        )
        raise Exception(msg)

    # cmap
    if cmap is None or vmin is None or vmax is None:
        if isinstance(coll.ddata[key]['data'], np.ndarray):
            nanmax = np.nanmax(coll.ddata[key]['data'])
            nanmin = np.nanmin(coll.ddata[key]['data'])
        else:
            nanmax = coll.ddata[key]['data'].max()
            nanmin = coll.ddata[key]['data'].min()
        diverging = nanmin * nanmax <= 0

    if cmap is None:
        if diverging:
            cmap = 'seismic'
        else:
            cmap = 'viridis'

    # vmin, vmax
    if vmin is None and diverging:
        vmin = -max(abs(nanmin), nanmax)
    if vmax is None and diverging:
        vmax = max(abs(nanmin), nanmax)

    # aspect
    aspect = _generic_check._check_var(
        aspect, 'aspect',
        default='equal',
        types=str,
        allowed=['auto', 'equal'],
    )

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = _generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = _generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return key, ind, cmap, vmin, vmax, aspect, dcolorbar, dleg


def plot_as_matrix(
    coll=None,
    key=None,
    ind=None,
    vmin=None,
    vmax=None,
    cmap=None,
    aspect=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
):

    # --------------
    # check input

    (
        key, ind, cmap, vmin, vmax, aspect, dcolorbar, dleg,
    ) = _plot_as_matrix_check(
        coll=coll,
        key=key,
        ind=ind,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        aspect=aspect,
        dcolorbar=dcolorbar,
        dleg=dleg,
    )

    # --------------
    #  Prepare data

    data = coll.ddata[key]['data']
    if hasattr(data, 'nnz'):
        data = data.toarray()
    n0, n1 = data.shape
    ref0, ref1 = coll.ddata[key]['ref']
    lab0 = f'ind0 ({ref0})'
    lab1 = f'ind1 ({ref1})'

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (12, 8)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.95,
                'bottom': 0.05, 'top': 0.90,
                'hspace': 0.15, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=3, nrows=3, **dmargin)

        ax0 = fig.add_subplot(gs[:2, :2], aspect='auto')
        ax0.set_ylabel(lab0)
        ax0.set_xlabel(lab1)
        ax0.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax0.xaxis.set_label_position('top')
        ax0.set_title(key, size=14, fontweight='bold')

        ax1 = fig.add_subplot(gs[:2, 2], sharey=ax0)
        ax1.set_xlabel('data')
        ax1.set_ylabel(lab0)
        ax1.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax1.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax1.yaxis.set_label_position('right')
        ax1.xaxis.set_label_position('top')

        ax2 = fig.add_subplot(gs[2, :2], sharex=ax0)
        ax2.set_ylabel('data')
        ax2.set_xlabel(lab1)

        dax = {
            'matrix': ax0,
            'misc1': {'ax': ax1, 'type': 'misc'},
            'misc2': {'ax': ax2, 'type': 'misc'},
        }

    dax = _generic_check._check_dax(dax=dax, main='matrix')

    # --------------
    # plot

    axtype = 'matrix'
    lkax = [kk for kk, vv in dax.items() if vv['type'] == axtype]
    for kax in lkax:
        ax = dax[kax]['ax']

        im = ax.imshow(
            data,
            extent=None,
            interpolation='nearest',
            origin='upper',
            aspect=aspect,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

        # plt.colorbar(im, ax=ax, **dcolorbar)

        # ind0, ind1
        ax.axhline(ind[0], c='k', lw=1., ls='-')
        ax.axvline(ind[1], c='k', lw=1., ls='-')

        if dleg is not False:
            ax.legend(**dleg)

    kax = 'misc1'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']

        ax.plot(
            data[:, ind[1]],
            np.arange(0, n0),
            ls='-',
            marker='.',
            lw=1.,
            color='k',
            label=f'ind0 = {ind[0]}',
        )

    kax = 'misc2'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']

        ax.plot(
            np.arange(0, n1),
            data[ind[0], :],
            ls='-',
            marker='.',
            lw=1.,
            color='k',
            label=f'ind1 = {ind[1]}',
        )
    return dax


# #############################################################################
# #############################################################################
#                       Spectral Lines
# #############################################################################


def _check_axvline_inputs(
    ymin=None, ymax=None,
    ls=None, lw=None, fontsize=None,
    side=None, fraction=None,
):

    if ymin is None:
        ymin = 0
    if ymax is None:
        ymax = 1
    if ls is None:
        ls = '-'
    if lw is None:
        lw = 1.
    if fontsize is None:
        fontsize = 9
    if side is None:
        side = 'right'
    if fraction is None:
        fraction = 0.75

    return ymin, ymax, ls, lw, fontsize, side, fraction


def _ax_axvline(
    ax=None, figsize=None, dmargin=None,
    quant=None, units=None, xlim=None,
    wintit=None, tit=None,
):

    if ax is None:

        if figsize is None:
            figsize = (9, 6)
        if dmargin is None:
            dmargin = {
                'left': 0.10, 'right': 0.90,
                'bottom': 0.10, 'top': 0.90,
                'hspace': 0.05, 'wspace': 0.05,
            }
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = ''

        fig = plt.figure(figsize=figsize)
        fig.canvas.manager.set_window_title(wintit)
        fig.suptitle(tit, size=12, fontweight='bold')

        gs = gridspec.GridSpec(1, 1, **dmargin)
        ax = fig.add_subplot(gs[0, 0])

        ax.set_ylim(0, 1)
        ax.set_xlim(xlim)
        ax.set_xlabel('{} ({})'.format(quant, units))

    return ax


def plot_axvline(
    din=None, key=None,
    sortby=None, dsize=None,
    param_x=None, param_txt=None,
    ax=None, ymin=None, ymax=None,
    ls=None, lw=None, fontsize=None,
    side=None, dcolor=None,
    fraction=None, units=None,
    figsize=None, dmargin=None,
    wintit=None, tit=None,
):

    # Check inputs
    ymin, ymax, ls, lw, fontsize, side, fraction = _check_axvline_inputs(
        ymin=ymin, ymax=ymax,
        ls=ls, lw=lw,
        fontsize=fontsize,
        side=side,
        fraction=fraction,
    )

    # Prepare data
    unique = sorted(set([din[k0][sortby] for k0 in key]))
    ny = len(unique)
    dy = (ymax-ymin)/ny
    ly = [(ymin+ii*dy, ymin+(ii+1)*dy) for ii in range(ny)]
    xside = 1.01 if side == 'right' else -0.01
    ha = 'left' if side == 'right' else 'right'

    if dcolor is None:
        lcol = plt.rcParams['axes.prop_cycle'].by_key()['color']
        dcolor = {uu: lcol[ii % len(lcol)] for ii, uu in enumerate(unique)}

    if dsize is not None:
        x, y = [], []
        colors = []
        sizes = []
        for ii, uu in enumerate(unique):
            lk = [
                k0 for k0 in key
                if din[k0][sortby] == uu and k0 in dsize.keys()
            ]
            if len(lk) > 0:
                x.append([din[k0][param_x] for k0 in lk])
                y.append([ly[ii][0]+fraction*dy/2. for k0 in lk])
                colors.append([dcolor[uu] for ii in range(len(lk))])
                sizes.append([dsize[k0] for k0 in lk])

        x = np.concatenate(x).ravel()
        y = np.concatenate(y).ravel()
        sizes = np.concatenate(sizes).ravel()
        colors = np.concatenate(colors).ravel()

    # plot preparation
    lamb = [din[k0][param_x] for k0 in key]
    Dlamb = np.nanmax(lamb) - np.nanmin(lamb)
    xlim = [np.nanmin(lamb) - 0.05*Dlamb, np.nanmax(lamb) + 0.05*Dlamb]
    ax = _ax_axvline(
        ax=ax, figsize=figsize, dmargin=dmargin,
        quant=param_x, units=units, xlim=xlim,
        wintit=wintit, tit=tit,
    )

    blend = transforms.blended_transform_factory(
        ax.transAxes, ax.transData
    )

    # plot
    for ii, uu in enumerate(unique):
        lk = [k0 for k0 in key if din[k0][sortby] == uu]
        for k0 in lk:
            ll = ax.axvline(
                x=din[k0][param_x],
                ymin=ly[ii][0],
                ymax=ly[ii][0] + fraction*dy,
                c=dcolor[uu],
                ls=ls,
                lw=lw,
            )

            ax.text(
                din[k0][param_x],
                ly[ii][1],
                din[k0][param_txt],
                color=dcolor[uu],
                horizontalalignment='center',
                verticalalignment='top',
                fontsize=fontsize,
                fontweight='normal',
                transform=ax.transData,
            )
        ax.text(
            xside,
            0.5*(ly[ii][0] + ly[ii][1]),
            uu,
            color=dcolor[uu],
            horizontalalignment=ha,
            verticalalignment='center',
            fontsize=fontsize+1,
            fontweight='bold',
            transform=blend,
        )

    # Add markers
    if dsize is not None:
        ax.scatter(
            x, y, s=sizes**2, c=colors,
            marker='o', edgecolors='None',
        )

    return ax


# #############################################################################
# #############################################################################
#               Dominance map
# #############################################################################


def _ax_dominance_map(
    dax=None, figsize=None, dmargin=None,
    x_scale=None, y_scale=None, amp_scale=None,
    quant=None, units=None,
    wintit=None, tit=None, dtit=None,
    proj=None, dlabel=None,
):

    allowed = ['map', 'spect', 'amp', 'prof']
    if dax is None:

        proj = _check_proj(proj=proj, allowed=allowed)
        dax = {}

        if figsize is None:
            figsize = (15, 9)
        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.90,
                'bottom': 0.08, 'top': 0.90,
                'hspace': 0.20, 'wspace': 0.50,
            }
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = ''
        if dtit is None:
            dtit = {}

        fig = plt.figure(figsize=figsize)
        fig.canvas.manager.set_window_title(wintit)
        fig.suptitle(tit, size=12, fontweight='bold')

        if len(proj) == 1:
            gs = gridspec.GridSpec(1, 1, **dmargin)
            for k0 in allowed:
                dax[k0] = fig.add_subplot(gs[0, 0])
        else:
            gs = gridspec.GridSpec(3, 5, **dmargin)
            shx, shy = None, None
            k0 = 'map'
            if k0 in proj:
                dax[k0] = fig.add_subplot(
                    gs[:, :2], xscale=x_scale, yscale=y_scale,
                )
            if 'spect' in proj:
                dax['spect'] = fig.add_subplot(
                    gs[0, 2:], yscale=amp_scale,
                )
                shy = dax['spect']
            if 'amp' in proj:
                dax['amp'] = fig.add_subplot(
                    gs[1, 2:], sharey=shy, yscale=amp_scale,
                )
                shx = dax['amp']
            if 'prof' in proj:
                dax['prof'] = fig.add_subplot(
                    gs[2, 2:], yscale=x_scale, sharex=shx,
                )

        for k0 in proj:
            if dtit is not None and dtit.get(k0) is not None:
                dax[k0].set_title(dtit[k0])

        for k0 in proj:
            if dlabel is not None and dlabel.get(k0) is not None:
                dax[k0].set_xlabel(dlabel[k0]['x'])
                dax[k0].set_ylabel(dlabel[k0]['y'])

    else:
        c0 = (
            isinstance(dax, dict)
            and all([ss in allowed for ss in dax.keys()])
        )
        if not c0:
            msg = (
                "\nArg dax must be a dict with the following allowed keys:\n"
                + "\t- allowed:  {}\n".format(allowed)
                + "\t- provided: {}".format(sorted(dax.keys()))
            )
            raise Exception(msg)

    return dax


def plot_dominance_map(
    din=None, key=None,
    im=None, extent=None,
    xval=None, yval=None, damp=None,
    x_scale=None, y_scale=None, amp_scale=None,
    sortby=None, dsize=None,
    param_x=None, param_txt=None,
    dax=None, proj=None,
    ls=None, lw=None, fontsize=None,
    side=None, dcolor=None,
    fraction=None, units=None,
    figsize=None, dmargin=None,
    wintit=None, tit=None, dtit=None, dlabel=None,
):

    # Check inputs

    # Prepare dax
    dax = _ax_dominance_map(
        dax=dax, proj=proj, figsize=figsize, dmargin=dmargin,
        quant=param_x, units=units,
        wintit=wintit, tit=tit, dtit=dtit, dlabel=dlabel,
        x_scale=x_scale, y_scale=y_scale, amp_scale=amp_scale,
    )

    if any([ss in dax.keys() for ss in ['spect', 'prof', 'amp']]):
        ind = np.arange(0, xval.size)

    k0 = 'map'
    if dax.get(k0) is not None:
        dax[k0].imshow(
            np.swapaxes(im, 0, 1),
            extent=extent,
            origin='lower',
            aspect='auto',
        )
        dax[k0].plot(xval, yval, ls='-', marker='.', lw=1., c='k')

    k0 = 'prof'
    if dax.get(k0) is not None:
        dax[k0].plot(ind, xval, ls='-', marker='.', lw=1., c='k')
        dax[k0].plot(ind, yval, ls='-', marker='.', lw=1., c='k')

    k0 = 'amp'
    if dax.get(k0) is not None:
        for k1 in damp.keys():
            dax[k0].plot(
                ind, damp[k1]['data'],
                ls='-', marker='.', lw=1., c=dcolor[damp[k1]['color']],
            )

    k0 = 'spect'
    if dax.get(k0) is not None:
        lamb = np.ones((list(damp.values())[0]['data'].size,))
        blend = transforms.blended_transform_factory(
            dax[k0].transData, dax[k0].transAxes
        )
        for k1 in damp.keys():
            lamb0 = din[k1]['lambda0']
            dax[k0].plot(
                lamb0*lamb, damp[k1]['data'],
                ls='None', marker='.', lw=1., c=dcolor[damp[k1]['color']],
            )
            dax[k0].text(
                lamb0,
                1.,
                din[k1][param_txt],
                color=dcolor[damp[k1]['color']],
                horizontalalignment='left',
                verticalalignment='bottom',
                fontsize=fontsize,
                fontweight='normal',
                transform=blend,
                rotation=60,
            )

        handles = [
            mlines.Line2D([], [], color=v0, label=k0)
            for k0, v0 in dcolor.items()
        ]
        dax[k0].legend(
            handles=handles,
            loc=2, bbox_to_anchor=(1., 1.),
        )

    return dax





# Built-in
import sys
import os
import warnings
import copy

# Common
import numpy as np
import scipy.interpolate
from scipy.interpolate import InterpolatedUnivariateSpline
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.axes._axes import Axes

# tofu
from tofu.version import __version__


# ##########################################################
# ##########################################################
#                  compute rocking curve
# ##########################################################
# ##########################################################


def compute_rockingcurve(
    ih=None, ik=None, il=None, lamb=None,
    use_non_parallelism=None, na=None,
    plot_asf=None, plot_power_ratio=None, plot_relation=None,
    verb=None, returnas=None,
):
    """The code evaluates, for a given wavelength and Miller indices set,
    the atomic plane distance d, the Bragg angle and the complex structure
    factor for alpha_Quartz crystals.
    Then, the power ratio and their integrated reflectivity, for 3 differents
    models (perfect, mosaic and dynamical) are computed to obtain
    rocking curves, for parallel and perpendicular photon polarizations.

    The possibility to add a non-parallelism between crystal's optical surface
    and inter-atomic planes is now available. Rocking curve plots are updated
    in order to show 3 cases: non-parallelism equal to zero and both limit
    cases close to +/- the reference Bragg angle.
    The relation between the value of this non-parallelism and 3 physical
    quantities can also be plotted: the integrated reflectivity for the normal
    component of polarization, the rocking curve widths and the symmetry
    parameter b.

    The alpha-Quartz, symmetry group D(4,3), is assumed left-handed.
    It makes a difference for certain planes, for example between
    (h,k,l)=(2,0,3) and (2,0,-3).
    The alpha-Quartz structure is hexagonal with 3 molecules of SiO2 in the
    unit cell.

    All crystal lattice constants and wavelengths are in Angstroms (1e-10 m).

    Parameters:
    -----------
    ih, ik, il:    int
        Miller indices of crystal used
    lamb:    float
        Wavelength of interest, in Angstroms (1e-10 m)
    use_non_parallelism:    str
        Introduce non-parallelism between dioptre and reflecting planes
    na:    int
        Number of non-parallelism angles steps, odd number preferred
    plot_asf:    str
        Plotting the atomic scattering factor thanks to data with respect to
        sin(theta)/lambda
    plot_power_ratio:    str
        Plot the power ratio with respect to the glancing angle
    plot_relation:    str
        Plot relations between the integrated reflectivity, the intrinsic
        width and the parameter b vs the glancing angle
    verb:    str
        True or False to print the content of the results dictionnary 'dout'
    returnas:    str
        Entry 'dict' to allow optionnal returning of 'dout' dictionnary
    """

    # Check inputs
    # ------------

    if use_non_parallelism is None:
        use_non_parallelism = False
    if na is None:
        na = 51
    if plot_asf is None:
        plot_asf = False
    if plot_power_ratio is None:
        plot_power_ratio = True
    if plot_relation is None and use_non_parallelism is not False:
        plot_relation = True
    if verb is None:
        verb = True
    if returnas is None:
        returnas = None

    ih, ik, il, lamb = CrystBragg_check_inputs_rockingcurve(
        ih=ih, ik=ik, il=il, lamb=lamb,
    )

    # Calculations of main crystal parameters
    # ---------------------------------------

    # Classical electronical radius, in Angstroms
    re = 2.8208e-5

    # Inter-atomic distances into hexagonal cell unit and associated volume
    # TBC with F.
    a0 = 4.9134
    c0 = 5.4051
    V = (a0**2.)*c0*np.sqrt(3.)/2.

    # Atomic number of Si and O atoms
    Zsi = 14.
    Zo = 8.

    # Position of the three Si atoms in the unit cell
    u = 0.4705
    xsi = np.r_[-u, u, 0.]
    ysi = np.r_[-u, 0., u]
    zsi = np.r_[1./3., 0., 2./3.]
    Nsi = np.size(xsi)

    # Position of the six O atoms in the unit cell
    x = 0.4152
    y = 0.2678
    z = 0.1184
    xo = np.r_[x, y - x, -y, x - y, y, -x]
    yo = np.r_[y, -x, x - y, -y, x, y - x]
    zo = np.r_[z, z + 1./3., z + 2./3., -z, 2./3. - z, 1./3. - z]
    No = np.size(xo)

    # Bragg angle and atomic plane distance d for a given wavelength lamb and
    # Miller indices (h,k,l)
    d_num = np.sqrt(3.)*a0*c0
    d_den = np.sqrt(4.*(ih**2 + ik**2 + ih*ik)*(c0**2) + 3.*(il**2)*(a0**2))
    if d_den == 0.:
        msg = (
            "Something went wrong in the calculation of d, equal to 0!\n"
            "Please verify the values for the following Miller indices:\n"
            + "\t - h: first Miller index ({})\n".format(ih)
            + "\t - k: second Miller index ({})\n".format(ik)
            + "\t - l: third Miller index ({})\n".format(il)
        )
        raise Exception(msg)
    d_atom = d_num/d_den
    if d_atom < lamb/2.:
        msg = (
            "According to Bragg law, Bragg scattering need d > lamb/2!\n"
            "Please check your wavelength argument.\n"
        )
        raise Exception(msg)
    sol = 1./(2.*d_atom)
    sin_theta = lamb/(2.*d_atom)
    theta = np.arcsin(sin_theta)
    theta_deg = theta*(180./np.pi)
    lc = [theta_deg < 10., theta_deg > 89.]
    if any(lc):
        msg = (
            "The computed value of theta is behind the arbitrary limits.\n"
            "Limit condition: 10° < theta < 89° and\n"
            "theta = ({})°\n".format(theta_deg)
        )
        raise Exception(msg)

    # Atomic scattering factors ("asf") for Si(2+) and O(1-) as a function of
    # sin(theta)/lambda ("sol"), taking into account molecular bounds
    # ("si") for Silicium and ("o") for Oxygen
    sol_si = np.r_[
        0., 0.1, 0.2, 0.25, 0.3, 0.35, 0.4, 0.5, 0.6,
        0.7, 0.8, 0.9, 1., 1.1, 1.2, 1.3, 1.4, 1.5,
    ]
    asf_si = np.r_[
        12., 11., 9.5, 8.8, 8.3, 7.7, 7.27, 6.25, 5.3,
        4.45, 3.75, 3.15, 2.7, 2.35, 2.07, 1.87, 1.71, 1.6,
    ]
    sol_o = np.r_[
        0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1., 1.1,
    ]
    asf_o = np.r_[
        9., 7.836, 5.756, 4.068, 2.968, 2.313, 1.934, 1.710, 1.566, 1.462,
        1.373, 1.294,
    ]


    # atomic absorption coefficient for Si and O as a function of lamb
    mu_si = 1.38e-2*(lamb**2.79)*(Zsi**2.73)
    mu_si1 = 5.33e-4*(lamb**2.74)*(Zsi**3.03)
    if lamb > 6.74:
        mu_si = mu_si1
    mu_o = 5.4e-3*(lamb**2.92)*(Zo**3.07)
    mu = 2.65e-8*(7.*mu_si + 8.*mu_o)/15.

    # Calculation of the structure factor for the alpha-quartz crystal
    # ----------------------------------------------------------------

    # interpolation of atomic scattering factor ("f") in function of sol
    # ("_re") for Real part and ("_im") for Imaginary part
    fsi_re = scipy.interpolate.interp1d(sol_si, asf_si)
    dfsi_re = 0.1335*lamb - 0.006
    fsi_re = fsi_re(sol) + dfsi_re
    fsi_im = 5.936e-4*Zsi*(mu_si/lamb)

    fo_re = scipy.interpolate.interp1d(sol_o, asf_o)
    dfo_re = 0.1335*lamb - 0.206
    fo_re = fo_re(sol) + dfo_re
    fo_im = 5.936e-4*Zo*(mu_o/lamb)

    # structure factor ("F") for (hkl) reflection
    phasesi = np.full((xsi.size), np.nan)
    phaseo = np.full((xo.size), np.nan)
    for i in range(xsi.size):
        phasesi[i] = ih*xsi[i] + ik*ysi[i] + il*zsi[i]
    for j in range(xo.size):
        phaseo[j] = ih*xo[j] + ik*yo[j] + il*zo[j]

    Fsi_re1 = np.sum(fsi_re*np.cos(2*np.pi*phasesi))
    Fsi_re2 = np.sum(fsi_re*np.sin(2*np.pi*phasesi))
    Fsi_im1 = np.sum(fsi_im*np.cos(2*np.pi*phasesi))
    Fsi_im2 = np.sum(fsi_im*np.sin(2*np.pi*phasesi))

    Fo_re1 = np.sum(fo_re*np.cos(2*np.pi*phaseo))
    Fo_re2 = np.sum(fo_re*np.sin(2*np.pi*phaseo))
    Fo_im1 = np.sum(fo_im*np.cos(2*np.pi*phaseo))
    Fo_im2 = np.sum(fo_im*np.sin(2*np.pi*phaseo))

    F_re_cos = Fsi_re1 + Fo_re1
    F_re_sin = Fsi_re2 + Fo_re2
    F_im_cos = Fsi_im1 + Fo_im1
    F_im_sin = Fsi_im2 + Fo_im2

    F_re = np.sqrt(F_re_cos**2 + F_re_sin**2)
    F_im = np.sqrt(F_im_cos**2 + F_im_sin**2)

    # Calculation of Fourier coefficients of polarization
    # ---------------------------------------------------

    # expression of the Fourier coef. psi_H
    Fmod = np.sqrt(
        F_re**2 + F_im**2 - 2.*(F_re_cos*F_re_sin - F_im_cos*F_im_sin)
    )

    # psi_-H equivalent to (-ih, -ik, -il)
    Fbmod = np.sqrt(
        F_re**2 + F_im**2 - 2.*(F_im_cos*F_im_sin - F_re_cos*F_re_sin)
    )

    if Fmod == 0.:
        Fmod == 1e-30
    if Fbmod == 0.:
        Fbmod == 1e-30

    # ratio imaginary part and real part of the structure factor
    kk = F_im/F_re

    # rek : Real(kk)
    rek = (F_re_cos*F_im_cos + F_re_sin*F_im_sin)/(F_re**2.)

    # real part of psi_H
    psi_re = (re*(lamb**2)*F_re)/(np.pi*V)

    # zero-order real part (averaged)
    # TBC with F.
    psi0_dre = -re*(lamb**2)*(
        No*(Zo + dfo_re) + Nsi*(Zsi + dfsi_re)
    )/(np.pi*V)
    """psi0_dre = -re*(lamb**2)*(
        No*Zo*(1. + dfo_re) + Nsi*Zsi*(1. + dfsi_re)
    )/(np.pi*V)"""

    # zero-order imaginary part (averaged)
    psi0_im = -re*(lamb**2)*(No*fo_im + Nsi*fsi_im)/(np.pi*V)
    """psi0_im = -re*(lamb**2)*(No*Zo*fo_im + Nsi*Zsi*fsi_im)/(np.pi*V)"""

    # Power ratio and their integrated reflectivity for 3 crystals models:
    # perfect (Darwin model), ideally mosaic thick and dynamical
    # --------------------------------------------------------------------

    if use_non_parallelism:
        (
            alpha, bb, polar, g, y, y0, power_ratio, th,
            rhg, rhg_perp_norm, rhg_para_norm,
            P_per, P_mos, P_dyn, det, det_norm,
        ) = CrystBragg_comp_integrated_reflect(
            lamb=lamb, re=re, V=V, Zo=Zo, theta=theta, mu=mu,
            F_re=F_re, psi_re=psi_re, psi0_dre=psi0_dre, psi0_im=psi0_im,
            Fmod=Fmod, Fbmod=Fbmod, kk=kk, rek=rek,
            model=['perfect', 'mosaic', 'dynamical'],
            use_non_parallelism=use_non_parallelism, na=na,
        )
    else:
        (
            alpha, bb, polar, g, y, y0, power_ratio, th,
            rhg, P_per, P_mos, P_dyn, det,
        ) = CrystBragg_comp_integrated_reflect(
            lamb=lamb, re=re, V=V, Zo=Zo, theta=theta, mu=mu,
            F_re=F_re, psi_re=psi_re, psi0_dre=psi0_dre, psi0_im=psi0_im,
            Fmod=Fmod, Fbmod=Fbmod, kk=kk, rek=rek,
            model=['perfect', 'mosaic', 'dynamical'],
            use_non_parallelism=use_non_parallelism, na=na,
        )

    # Plot atomic scattering factor
    # -----------------------------

    if plot_asf:
        CrystalBragg_plot_atomic_scattering_factor(
            sol_si=sol_si, sol_o=sol_o, asf_si=asf_si, asf_o=asf_o,
        )

    # Plot power ratio
    # ----------------

    if plot_power_ratio:
        CrystalBragg_plot_power_ratio(
            ih=ih, ik=ik, il=il, lamb=lamb,
            theta=theta, theta_deg=theta_deg,
            th=th, power_ratio=power_ratio, y=y, y0=y0,
            bb=bb, polar=polar, alpha=alpha,
            use_non_parallelism=use_non_parallelism, na=na,
        )

    # Plot P_dyn vs glancing angle
    # ----------------------------

    if plot_relation:
        CrystalBragg_plot_reflect_glancing(
            ih=ih, ik=ik, il=il, lamb=lamb,
            theta=theta, theta_deg=theta_deg,
            alpha=alpha, bb=bb, th=th, rhg=rhg,
            rhg_perp_norm=rhg_perp_norm,
            rhg_para_norm=rhg_para_norm,
            det=det_norm,
        )

    # Print results
    # -------------

    dout = {
        'Wavelength (A)': lamb,
        'Miller indices': (ih, ik, il),
        'Inter-reticular distance (A)': d_atom,
        'Volume of the unit cell (A^3)': V,
        'Bragg angle of reference (rad, deg)': (theta, theta_deg),
        'Ratio imag & real part of structure factor': kk,
        'Integrated reflectivity': {
            'perfect model': P_per,
            'mosaic model': P_mos,
            'dynamical model': P_dyn,
        },
        'P_{dyn,para}/P_{dyn,norm} (integrated values)': rhg[1]/rhg[0],
        'RC width': det,
    }
    if use_non_parallelism:
        dout['Non-parallelism angles (deg)'] = alpha*(180/np.pi)

    if verb is True:
        dout['Inter-reticular distance (A)'] = np.round(d_atom, decimals=3)
        dout['Volume of the unit cell (A^3)'] = np.round(V, decimals=3)
        dout['Bragg angle of reference (rad, deg)'] = (
            np.round(theta, decimals=3), np.round(theta_deg, decimals=3),
        )
        dout['Ratio imag & real part of structure factor'] = (
            np.round(kk, decimals=3,)
        )
        dout['P_{dyn,para}/P_{dyn,norm} (integrated values)'] = np.round(
            rhg[1]/rhg[0], decimals=9,
        )
        dout['RC width'] = np.round(det, decimals=6)
        dout['Integrated reflectivity']['perfect model'] = (
            np.round(P_per, decimals=9),
        )
        dout['Integrated reflectivity']['mosaic model'] = (
            np.round(P_mos, decimals=9),
        )
        dout['Integrated reflectivity']['dynamical model'] = (
            np.round(P_dyn, decimals=9),
        )
        lstr = [f'\t -{k0}: {V0}' for k0, V0 in dout.items()]
        msg = (
            " The following data was calculated:\n"
            + "\n".join(lstr)
        )
        print(msg)

    if returnas is dict:
        return dout


def CrystBragg_check_inputs_rockingcurve(
    ih=None, ik=None, il=None, lamb=None,
):

    # All args are None
    dd = {'ih': ih, 'ik': ik, 'il': il, 'lamb': lamb}
    lc = [v0 is None for k0, v0 in dd.items()]
    if all(lc):
        ih = 1
        ik = 1
        il = 0
        lamb = 3.96
        msg = (
            "Args h, k, l and lamb were not explicitely specified\n"
            "and have been put to the following default values:\n"
            + "\t - h: first Miller index ({})\n".format(ih)
            + "\t - k: second Miller index ({})\n".format(ik)
            + "\t - l: third Miller index ({})\n".format(il)
            + "\t - lamb: wavelength of interest ({})\n".format(lamb)
        )
        warnings.warn(msg)

    # Some args are bot but not all
    dd2 = {'ih': ih, 'ik': ik, 'il': il, 'lamb': lamb}
    lc2 = [v0 is None for k0, v0 in dd2.items()]
    if any(lc2):
        msg = (
            "Args h, k, l and lamb must be provided together:\n"
            + "\t - h: first Miller index ({})\n".format(ih)
            + "\t - k: second Miller index ({})\n".format(ik)
            + "\t - l: third Miller index ({})\n".format(il)
            + "\t - lamb: wavelength of interest ({})\n".format(lamb)
        )
        raise Exception(msg)

    # Some args are string values
    cdt = [type(v0) == str for k0, v0 in dd.items()]
    if any(cdt) or all(cdt):
        msg = (
            "Args h, k, l and lamb must not be string inputs:\n"
            "and have been put to default values:\n"
            + "\t - h: first Miller index ({})\n".format(ih)
            + "\t - k: second Miller index ({})\n".format(ik)
            + "\t - l: third Miller index ({})\n".format(il)
            + "\t - lamb: wavelength of interest ({})\n".format(lamb)
        )
        raise Exception(msg)

    return ih, ik, il, lamb,


def CrystBragg_comp_integrated_reflect(
    lamb=None, re=None, V=None, Zo=None, theta=None, mu=None,
    F_re=None, psi_re=None, psi0_dre=None, psi0_im=None,
    Fmod=None, Fbmod=None, kk=None, rek=None,
    model=[None, None, None],
    use_non_parallelism=None, na=None,
):

    # Symmetry parameter b
    # ---------------------

    if not use_non_parallelism:
        alpha = np.r_[0.]
        bb = np.r_[-1.]
    else:
        alpha = np.linspace(-theta + 0.01, theta - 0.01, na)
        # alpha = np.linspace(-0.05, 0.05, 5)*(np.pi/180)
        bb = np.sin(alpha + theta)/np.sin(alpha - theta)

    # Perfect (darwin) model
    # ----------------------

    P_per = Zo*F_re*re*(lamb**2)*(1. + abs(np.cos(2.*theta)))/(
        6.*np.pi*V*np.sin(2.*theta)
    )

    # Ideally thick mosaic model
    # --------------------------

    P_mos = (F_re**2)*(re**2)*(lamb**3)*(1. + (np.cos(2.*theta))**2)/(
        4.*mu*(V**2)*np.sin(2.*theta)
    )

    # Dynamical model
    # ---------------

    # incident wave polarization (normal & parallel components)
    polar = np.r_[1., abs(np.cos(2.*theta))]

    # variables of simplification y, dy, g, L
    g = np.full((polar.size, bb.size), np.nan)
    for i in range(polar.size):
        for j in range(bb.size):
            g[i, j] = (((1. - bb[j])/2.)*psi0_im)/(
                np.sqrt(abs(bb[j]))*polar[i]*psi_re
            )
    y = np.linspace(-10., 10., 201)
    dy = np.zeros(201) + 0.1
    y0 = -psi0_dre/np.sin(2.*theta)

    al = np.full((polar.size, bb.size, y.size), 0.)
    power_ratio = np.full((al.shape), np.nan)
    power_ratiob = np.full((bb.size, y.size), np.nan)
    rhy = np.full((bb.size), np.nan)
    th = np.full((al.shape), np.nan)
    conv_ygscale = np.full((polar.size, bb.size), np.nan)
    rhg = np.full((polar.size, bb.size), np.nan)

    for i in range(polar.size):
        for j in range(bb.size):
            al[i, j, ...] = (y**2 + g[i, j]**2 + np.sqrt(
                (y**2 - g[i, j]**2 + abs(kk)**2 - 1.)**2 + 4.*(
                    g[i, j]*y - rek
                )**2
            ))/np.sqrt((abs(kk)**2 - 1.)**2 + 4.*(rek**2))
            # reflecting power or power ratio R_dyn
            power_ratio[i, j, ...] = (Fmod/Fbmod)*(
                al[i, j, :] - np.sqrt((al[i, j, :]**2) - 1.)
            )
            # integration of the power ratio over dy
            power_ratiob[j, :] = power_ratio[i, j, ...]
            rhy[j] = np.sum(dy*power_ratiob[j, :])
            # conversion formula from y-scale to glancing angle scale
            th[i, j, ...] = (
                -y*polar[i]*psi_re*np.sqrt(abs(bb[j])) + psi0_dre*(
                    (1. - bb[j])/2.
                )
            )/(bb[j]*np.sin(2.*theta))
            # integrated reflecting power in the glancing angle scale
            # r(i=0): normal component & r(i=1): parallel component
            conv_ygscale[i, ...] = (polar[i]*psi_re)/(
                np.sqrt(abs(bb))*np.sin(2*theta)
            )
            rhg[i, ...] = conv_ygscale[i, :]*rhy

    # Integrated reflectivity and rocking curve width
    def lin_interp(x, y, i, half):
        return x[i] + (x[i+1] - x[i])*((half - y[i])/(y[i+1] - y[i]))

    def half_max_x(x, y):
        half = np.max(y)/2.
        signs = np.sign(np.add(y, -half))
        zero_cross = (signs[0:-2] != signs[1:-1])
        zero_cross_ind = np.where(zero_cross)[0]
        return [
            lin_interp(x, y, zero_cross_ind[0], half),
            lin_interp(x, y, zero_cross_ind[1], half),
        ]

    P_dyn = np.full((bb.size), np.nan)
    rhg_perp = P_dyn.copy()
    rhg_para = P_dyn.copy()
    det = P_dyn.copy()

    for j in range(bb.size):
        rhg_perp[j] = rhg[0, j]
        rhg_para[j] = rhg[1, j]
        # each component accounts for half the intensity of the incident beam
        # if not polarized: reflecting power is an average over the 2 states
        P_dyn[j] = np.sum(rhg[:, j])/2.
        if P_dyn[j] < 1e-7:
            msg = (
                "Please check the equations for integrated reflectivity:\n"
                "the value of P_dyn ({}) is less than 1e-7.\n".format(P_dyn[j])
            )
            raise Exception(msg)
        hmx = half_max_x(th[0, j, :], power_ratio[0, j, :])
        det[j] = hmx[1] - hmx[0]

    # Normalization for alpha=0 case
    if use_non_parallelism:
        rhg_perp_norm = np.full((rhg_perp.size), np.nan)
        rhg_para_norm = np.full((rhg_para.size), np.nan)
        det_norm = np.full((det.size), np.nan)
        nn = (det.size/2.)
        if (nn % 2) == 0:
            nn = int(nn - 1)
        else:
            nn = int(nn - 0.5)
        det_norm = det/det[nn]
        rhg_perp_norm = rhg_perp/rhg_perp[nn]
        rhg_para_norm = rhg_para/rhg_para[nn]

    if use_non_parallelism:
        return (
            alpha, bb, polar, g, y, y0, power_ratio, th,
            rhg, rhg_perp_norm, rhg_para_norm,
            P_per, P_mos, P_dyn, det, det_norm,
        )
    else:
        return (
            alpha, bb, polar, g, y, y0, power_ratio, th,
            rhg, P_per, P_mos, P_dyn, det,
        )


def CrystalBragg_plot_atomic_scattering_factor(
    sol_si=None, sol_o=None, asf_si=None, asf_o=None,
):

    # Check inputs
    # ------------

    lc = [sol_si is None, sol_o is None, asf_si is None, asf_o is None]
    if any(lc):
        msg = (
            "Please make sure that all entry arguments are valid and not None!"
        )
        raise Exception(msg)

    # Plot
    # ----

    fig = plt.figure(figsize=(8, 6))
    gs = gridspec.GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r'sin($\theta$)/$\lambda$')
    ax.set_ylabel("atomic scattering factor")
    ax.plot(sol_si, asf_si, label="Si")
    ax.plot(sol_o, asf_o, label="O")
    ax.legend()


def CrystalBragg_plot_power_ratio(
    ih=None, ik=None, il=None, lamb=None,
    theta=None, theta_deg=None,
    th=None, power_ratio=None, y=None, y0=None,
    bb=None, polar=None, alpha=None,
    use_non_parallelism=None, na=None,
):

    # Plot
    # ----

    fig1 = plt.figure(figsize=(8, 6))
    gs = gridspec.GridSpec(1, 1)
    ax = fig1.add_subplot(gs[0, 0])
    ax.set_title(
        'Qz, ' + f'({ih},{ik},{il})' + fr', $\lambda$={lamb} A' +
        fr', Bragg angle={np.round(theta_deg, decimals=3)}$\deg$'
    )
    ax.set_xlabel(r'$\theta$-$\theta_{B}$ (rad)')
    ax.set_ylabel('P$_H$/P$_0$')

    if use_non_parallelism:
        alpha_deg = alpha*(180/np.pi)
        nalpha = alpha_deg.size
        if (nalpha % 2) == 0:
            nalpha = int(nalpha - 1)
        else:
            nalpha = int(nalpha - 0.5)
        dd = np.r_[0, int(nalpha/2), nalpha]
        let = {'I': dd[0], 'II': dd[1], 'III': dd[2]}
    else:
        alpha_deg = np.r_[alpha]*(180/np.pi)
        dd = np.r_[0]
        let = {'I': dd[0]}

    for j in range(na):
        if any(j == dd):
            ind = np.where(
                power_ratio[0, j, :] == np.amax(power_ratio[0, j, :])
            )
            keylist = list(let.keys())
            valuelist = list(let.values())
            valuedd = valuelist.index(j)
            keydd = keylist[valuedd]
            ax.text(
                th[0, j, ind],
                np.max(power_ratio[0, j, :] + 0.005),
                '({})'.format(keydd),
            )
            ax.plot(
                th[0, j, :],
                power_ratio[0, j, :],
                'k-',
                label=r'normal, ({}): $\alpha$=({})deg'.format(
                    keydd, np.round(alpha_deg[j], 3)
                ),
            )
            ax.plot(
                th[1, j, :],
                power_ratio[1, j, :],
                'k:',
                label=r'parallel',
            )
    # ax.axvline(y0, linetsyle=":", label='pattern centeri in y-scale')
    ax.legend()


def CrystalBragg_plot_reflect_glancing(
    ih=None, ik=None, il=None, lamb=None, theta=None, theta_deg=None,
    alpha=None, bb=None, th=None,
    rhg=None, rhg_perp_norm=None, rhg_para_norm=None, det=None,
):

    # Plot
    # ----

    fig2 = plt.figure(figsize=(8, 6))
    gs = gridspec.GridSpec(1, 1)
    ax = fig2.add_subplot(gs[0, 0])
    ax.set_title(
        'Qz, ' + f'({ih},{ik},{il})' + fr', $\lambda$={lamb} A' +
        fr', Bragg angle={np.round(theta_deg, decimals=3)}deg'
    )
    ax.set_xlabel(r'$\alpha$ (deg)')
    ax.set_xlim(-theta_deg - 10., theta_deg + 10.)
    ax.set_ylim(0., 5.)

    alpha_deg = alpha*(180/np.pi)
    f = scipy.interpolate.interp1d(alpha_deg, abs(bb), kind='cubic')
    alpha_deg_bis = np.linspace(-alpha_deg, alpha_deg, 21)
    bb_bis = f(alpha_deg_bis)
    ax.plot(
        alpha_deg,
        det,
        'k:',
        label=r'RC width $\Delta\theta$ (normalized)',
    )
    ax.plot(
        alpha_deg,
        rhg_perp_norm,
        'k-',
        label='P$_{dyn}$ (normal comp.) (normalized)',
    )
    ax.plot(
        alpha_deg,
        rhg_para_norm,
        'k--',
        label='P$_{dyn}$ (parallel comp.) (normalized)',
    )
    ax.plot(
        alpha_deg_bis[:, 0],
        bb_bis[:, 0],
        'k-.',
        label='|b|',
    )
    ax.legend()




# -*- coding: utf-8 -*-


# common
import matplotlib.pyplot as plt


_LALLOWED_AXESTYPES = [
    'cross', 'hor',
    'matrix',
    'timetrace',
    'profile1d',
    'image',
    'misc'
]


# #############################################################################
# #############################################################################
#                           Utilities
# #############################################################################


def _check_var(
    var,
    varname,
    types=None,
    default=None,
    allowed=None,
    excluded=None,
):

    # set to default
    if var is None:
        var = default
    if var is None and allowed is not None and len(allowed) == 1:
        var = allowed[0]

    # check type
    if types is not None:
        if not isinstance(var, types):
            msg = (
                f"Arg {varname} must be of type {types}!\n"
                f"Provided: {type(var)}"
            )
            raise Exception(msg)

    # check if allowed
    if allowed is not None:
        if var not in allowed:
            msg = (
                f"Arg {varname} must be in {allowed}!\n"
                f"Provided: {var}"
            )
            raise Exception(msg)

    # check if excluded
    if excluded is not None:
        if var in excluded:
            msg = (
                f"Arg {varname} must not be in {excluded}!\n"
                f"Provided: {var}"
            )
            raise Exception(msg)

    return var


# #############################################################################
# #############################################################################
#                   Utilities for plotting
# #############################################################################


def _check_dax(dax=None, main=None):

    # None
    if dax is None:
        return dax

    # Axes
    if issubclass(dax.__class__, plt.Axes):
        if main is None:
            msg = (
            )
            raise Exception(msg)
        else:
            return {main: dax}

    # dict
    c0 = (
        isinstance(dax, dict)
        and all([
            isinstance(k0, str)
            and (
                (
                    k0 in _LALLOWED_AXESTYPES
                    and issubclass(v0.__class__, plt.Axes)
                )
                or (
                    isinstance(v0, dict)
                    and issubclass(v0.get('ax').__class__, plt.Axes)
                    and v0.get('type') in _LALLOWED_AXESTYPES
                )
            )
            for k0, v0 in dax.items()
        ])
    )
    if not c0:
        msg = (
        )
        raise Exception(msg)

    for k0, v0 in dax.items():
        if issubclass(v0.__class__, plt.Axes):
            dax[k0] = {'ax': v0, 'type': k0}

    return dax



# -*- coding: utf-8 -*-


# Built-in
import sys
import os
# import itertools as itt
import copy
import warnings
from abc import ABCMeta, abstractmethod
import inspect


# Common
import numpy as np
import scipy.interpolate as scpinterp
# import matplotlib.pyplot as plt
# from matplotlib.tri import Triangulation as mplTri


# tofu
# from tofu import __version__ as __version__
import tofu.utils as utils
from . import _DataCollection_check_inputs
from . import _comp
from . import _DataCollection_comp
from . import _DataCollection_plot
from . import _def
from . import _comp_spectrallines


__all__ = ['DataCollection']    # , 'TimeTraceCollection']


_INTERPT = 'zero'
_GROUP_0D = 'time'
_GROUP_1D = 'radius'
_GROUP_2D = 'mesh2d'


#############################################
#############################################
#       Abstract Parent class
#############################################
#############################################


class DataCollection(utils.ToFuObject):
    """ A generic class for handling data

    Provides methods for:
        - introspection
        - plateaux finding
        - visualization

    """
    __metaclass__ = ABCMeta

    # Fixed (class-wise) dictionary of default properties
    _ddef = {
        'Id': {'include': ['Mod', 'Cls', 'Name', 'version']},
        'params': {
            'ddata': {
                'source': (str, 'unknown'),
                'dim':    (str, 'unknown'),
                'quant':  (str, 'unknown'),
                'name':   (str, 'unknown'),
                'units':  (str, 'a.u.'),
            },
            'dobj': {},
         },
    }

    _forced_group = None
    if _forced_group is not None:
        _allowed_groups = [_forced_group]
    else:
        _allowed_groups = None
    # _dallowed_params = None
    _data_none = None
    _reserved_keys = None

    _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'
    _max_ndim = None

    _dgroup = {}
    _dref = {}
    _dref_static = {}
    _ddata = {}
    _dobj = {}

    _group0d = _GROUP_0D
    _group1d = _GROUP_1D
    _group2d = _GROUP_2D

    def __init_subclass__(cls, **kwdargs):
        # Does not exist before Python 3.6 !!!
        # Python 2
        super(DataCollection, cls).__init_subclass__(**kwdargs)
        # Python 3
        # super().__init_subclass__(**kwdargs)
        cls._ddef = copy.deepcopy(DataCollection._ddef)
        # cls._dplot = copy.deepcopy(Struct._dplot)
        # cls._set_color_ddef(cls._color)

    def __init__(
        self,
        dgroup=None,
        dref=None,
        dref_static=None,
        ddata=None,
        dobj=None,
        Id=None,
        Name=None,
        fromdict=None,
        SavePath=None,
        include=None,
        sep=None,
    ):

        # Create a dplot at instance level
        # self._dplot = copy.deepcopy(self.__class__._dplot)
        kwdargs = locals()
        del kwdargs['self']
        super().__init__(**kwdargs)

    def _reset(self):
        # Run by the parent class __init__()
        super()._reset()
        self._dgroup = {}
        self._dref = {}
        self._dref_static = {}
        self._ddata = {}
        self._dobj = {}

    @classmethod
    def _checkformat_inputs_Id(cls, Id=None, Name=None,
                               include=None, **kwdargs):
        if Id is not None:
            assert isinstance(Id, utils.ID)
            Name = Id.Name
        # assert isinstance(Name, str), Name
        if include is None:
            include = cls._ddef['Id']['include']
        kwdargs.update({'Name': Name, 'include': include})
        return kwdargs

    ###########
    # Get check and format inputs
    ###########

    ###########
    # _init
    ###########

    def _init(
        self,
        dgroup=None,
        dref=None,
        dref_static=None,
        ddata=None,
        dobj=None,
        **kwargs,
    ):
        self.update(
            dgroup=dgroup,
            dref=dref,
            dref_static=dref_static,
            ddata=ddata,
            dobj=dobj,
        )
        self._dstrip['strip'] = 0

    ###########
    # set dictionaries
    ###########

    def update(
        self,
        dobj=None,
        ddata=None,
        dref=None,
        dref_static=None,
        dgroup=None,
    ):
        """ Can be used to set/add data/ref/group

        Will update existing attribute with new dict
        """
        # Check consistency
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs._consistency(
                    dobj=dobj, dobj0=self._dobj,
                    ddata=ddata, ddata0=self._ddata,
                    dref=dref, dref0=self._dref,
                    dref_static=dref_static, dref_static0=self._dref_static,
                    dgroup=dgroup, dgroup0=self._dgroup,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params']['ddata'],
                    ddefparams_obj=self._ddef['params']['dobj'],
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    # ---------------------
    # Adding group / ref / quantity one by one
    # ---------------------

    def add_group(self, group=None):
        # Check consistency
        self.update(ddata=None, dref=None, dref_static=None, dgroup=group)

    def add_ref(self, key=None, group=None, data=None, **kwdargs):
        dref = {key: {'group': group, 'data': data, **kwdargs}}
        # Check consistency
        self.update(ddata=None, dref=dref, dref_static=None, dgroup=None)

    # TBF
    def add_ref_static(self, key=None, which=None, **kwdargs):
        dref_static = {which: {key: kwdargs}}
        # Check consistency
        self.update(
            ddata=None, dref=None, dref_static=dref_static, dgroup=None,
        )

    def add_data(self, key=None, data=None, ref=None, **kwdargs):
        ddata = {key: {'data': data, 'ref': ref, **kwdargs}}
        # Check consistency
        self.update(ddata=ddata, dref=None, dref_static=None, dgroup=None)

    def add_obj(self, which=None, key=None, **kwdargs):
        dobj = {which: {key: kwdargs}}
        # Check consistency
        self.update(dobj=dobj, dref=None, dref_static=None, dgroup=None)

    # ---------------------
    # Removing group / ref / quantities
    # ---------------------

    def remove_group(self, group=None):
        """ Remove a group (or list of groups) and all associated ref, data """
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs._remove_group(
                    group=group,
                    dgroup0=self._dgroup, dref0=self._dref, ddata0=self._ddata,
                    dref_static0=self._dref_static,
                    dobj0=self._dobj,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params']['ddata'],
                    ddefparams_obj=self._ddef['params']['dobj'],
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    def remove_ref(self, key=None, propagate=None):
        """ Remove a ref (or list of refs) and all associated data """
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs._remove_ref(
                    key=key,
                    dgroup0=self._dgroup, dref0=self._dref, ddata0=self._ddata,
                    dref_static0=self._dref_static,
                    dobj0=self._dobj,
                    propagate=propagate,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params']['ddata'],
                    ddefparams_obj=self._ddef['params']['dobj'],
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    def remove_ref_static(self, key=None, which=None, propagate=None):
        """ Remove a static ref (or list) or a whole category

        key os provided:
            => remove only the desired key(s)
                works only if key is not used in ddata and dobj

        which is provided:
            => treated as param, the whole category of ref_static is removed
                if propagate, the parameter is removed from ddata and dobj
        """
        _DataCollection_check_inputs._remove_ref_static(
            key=key,
            which=which,
            propagate=propagate,
            dref_static0=self._dref_static,
            ddata0=self._ddata,
            dobj0=self._dobj,
        )

    def remove_data(self, key=None, propagate=True):
        """ Remove a data (or list of data) """
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs._remove_data(
                    key=key,
                    dgroup0=self._dgroup, dref0=self._dref, ddata0=self._ddata,
                    dref_static0=self._dref_static,
                    dobj0=self._dobj,
                    propagate=propagate,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params']['ddata'],
                    ddefparams_obj=self._ddef['params']['dobj'],
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    def remove_obj(self, key=None, which=None, propagate=True):
        """ Remove a data (or list of data) """
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs._remove_obj(
                    key=key,
                    which=which,
                    dobj0=self._dobj,
                    ddata0=self._ddata,
                    dgroup0=self._dgroup,
                    dref0=self._dref,
                    dref_static0=self._dref_static,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params']['ddata'],
                    ddefparams_obj=self._ddef['params']['dobj'],
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    # ---------------------
    # Get / set / add / remove params
    # ---------------------

    def __check_which(self, which=None, return_dict=None):
        """ Check which in ['data'] + list(self._dobj.keys() """
        return _DataCollection_check_inputs._check_which(
            ddata=self._ddata,
            dobj=self._dobj,
            which=which,
            return_dict=return_dict,
        )

    def get_lparam(self, which=None):
        """ Return the list of params for the chosen dict ('data' or dobj[<>])
        """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return

        lp = list(list(dd.values())[0].keys())
        if which == 'data':
            lp.remove('data')
        return lp

    def get_param(
        self,
        param=None,
        key=None,
        ind=None,
        returnas=None,
        which=None,
    ):
        """ Return the array of the chosen parameter (or list of parameters)

        Can be returned as:
            - dict: {param0: {key0: values0, key1: value1...}, ...}
            - np[.ndarray: {param0: np.r_[values0, value1...], ...}

        """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        return _DataCollection_check_inputs._get_param(
            dd=dd, dd_name=which,
            param=param, key=key, ind=ind, returnas=returnas,
        )

    def set_param(
        self,
        param=None,
        value=None,
        ind=None,
        key=None,
        which=None,
    ):
        """ Set the value of a parameter

        value can be:
            - None
            - a unique value (int, float, bool, str, tuple) common to all keys
            - an iterable of vlues (array, list) => one for each key

        A subset of keys can be chosen (ind, key, fed to self.select()) to set
        only the value of some key

        """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        _DataCollection_check_inputs._set_param(
            dd=dd, dd_name=which,
            param=param, value=value, ind=ind, key=key,
        )

    def add_param(
        self,
        param,
        value=None,
        which=None,
    ):
        """ Add a parameter, optionnally also set its value """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        _DataCollection_check_inputs._add_param(
            dd=dd, dd_name=which,
            param=param, value=value,
        )

    def remove_param(
        self,
        param=None,
        which=None,
    ):
        """ Remove a parameter, none by default, all if param = 'all' """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        _DataCollection_check_inputs._remove_param(
            dd=dd, dd_name=which,
            param=param,
        )

    ###########
    # strip dictionaries
    ###########

    def _strip_ddata(self, strip=0, verb=0):
        pass

    ###########
    # _strip and get/from dict
    ###########

    @classmethod
    def _strip_init(cls):
        cls._dstrip['allowed'] = [0, 1]
        nMax = max(cls._dstrip['allowed'])
        doc = """
                 1: None
                 """
        doc = utils.ToFuObjectBase.strip.__doc__.format(doc, nMax)
        cls.strip.__doc__ = doc

    def strip(self, strip=0, verb=True):
        # super()
        super(DataCollection, self).strip(strip=strip, verb=verb)

    def _strip(self, strip=0, verb=True):
        self._strip_ddata(strip=strip, verb=verb)

    def _to_dict(self):
        dout = {
            'dgroup': {'dict': self._dgroup, 'lexcept': None},
            'dref': {'dict': self._dref, 'lexcept': None},
            'dref_static': {'dict': self._dref_static, 'lexcept': None},
            'ddata': {'dict': self._ddata, 'lexcept': None},
            'dobj': {'dict': self._dobj, 'lexcept': None},
        }
        return dout

    def _from_dict(self, fd):
        for k0 in ['dgroup', 'dref', 'ddata', 'dref_static', 'dobj']:
            if fd.get(k0) is not None:
                getattr(self, '_'+k0).update(**fd[k0])
        self.update()

    ###########
    # properties
    ###########

    @property
    def dgroup(self):
        """ The dict of groups """
        return self._dgroup

    @property
    def dref(self):
        """ the dict of references """
        return self._dref

    @property
    def dref_static(self):
        """ the dict of references """
        return self._dref_static

    @property
    def ddata(self):
        """ the dict of data """
        return self._ddata

    @property
    def dobj(self):
        """ the dict of obj """
        return self._dobj

    ###########
    # General use methods
    ###########

    def to_DataFrame(self, which=None):
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        import pandas as pd
        return pd.DataFrame(dd)

    # ---------------------
    # Key selection methods
    # ---------------------

    def select(self, which=None, log=None, returnas=None, **kwdargs):
        """ Return the indices / keys of data matching criteria

        The selection is done comparing the value of all provided parameters
        The result is a boolean indices array, optionally with the keys list
        It can include:
            - log = 'all': only the data matching all criteria
            - log = 'any': the data matching any criterion

        If log = 'raw', a dict of indices arrays is returned, showing the
        details for each criterion

        """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        return _DataCollection_check_inputs._select(
            dd=dd, dd_name=which,
            log=log, returnas=returnas, **kwdargs,
        )

    def _ind_tofrom_key(
        self,
        ind=None,
        key=None,
        group=None,
        returnas=int,
        which=None,
    ):
        """ Return ind from key or key from ind for all data """
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        return _DataCollection_check_inputs._ind_tofrom_key(
            dd=dd, dd_name=which, ind=ind, key=key,
            group=group, dgroup=self._dgroup,
            returnas=returnas,
        )

    def _get_sort_index(self, which=None, param=None):
        """ Return sorting index ofself.ddata dict """

        if param is None:
            return

        if param == 'key':
            ind = np.argsort(list(dd.keys()))
        elif isinstance(param, str):
            ind = np.argsort(
                self.get_param(param, which=which, returnas=np.ndarray)[param]
            )
        else:
            msg = "Arg param must be a valid str\n  Provided: {}".format(param)
            raise Exception(msg)
        return ind

    def sortby(self, param=None, order=None, which=None):
        """ sort the self.ddata dict by desired parameter """

        # Trivial case
        if len(self._ddata) == 0 and len(self._dobj) == 0:
            return

        # --------------
        # Check inputs

        # order
        if order is None:
            order = 'increasing'

        c0 = order in ['increasing', 'reverse']
        if not c0:
            msg = (
                """
                Arg order must be in [None, 'increasing', 'reverse']
                Provided: {}
                """.format(order)
            )
            raise Exception(msg)

        # which
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return

        # --------------
        # sort
        ind = self._get_sort_index(param=param, which=which)
        if ind is None:
            return
        if order == 'reverse':
            ind = ind[::-1]

        lk = list(dd.keys())
        dd = {lk[ii]: dd[lk[ii]] for ii in ind}

        if which == 'data':
            self._ddata = dd
        else:
            self._dobj[which] = dd

    # ---------------------
    # Get refs from data key
    # ---------------------

    def _get_ref_from_key(self, key=None, group=None):
        """ Get the key of the ref in chosen group """

        # Check input
        if key not in self._ddata.keys():
            msg = "Provide a valid data key!\n\t- Provided: {}".format(key)
            raise Exception(msg)

        ref = self._ddata[key]['ref']
        if len(ref) > 1:
            if group not in self._dgroup.keys():
                msg = "Provided group is not valid!\n\t{}".format(group)
                raise Exception(msg)
            ref = [rr for rr in ref if self._dref[rr]['group'] == group]
            if len(ref) != 1:
                msg = "Ambiguous ref for key {}!\n\t- {}".format(key, ref)
                raise Exception(msg)
        return ref[0]

    # ---------------------
    # Switch ref
    # ---------------------

    def switch_ref(self, new_ref=None):
        """Use the provided key as ref (if valid) """
        self._dgroup, self._dref, self._dref_static, self._ddata, self._dobj =\
                _DataCollection_check_inputs.switch_ref(
                    new_ref=new_ref,
                    ddata=self._ddata,
                    dref=self._dref,
                    dgroup=self._dgroup,
                    dobj0=self._dobj,
                    dref_static0=self._dref_static,
                    allowed_groups=self._allowed_groups,
                    reserved_keys=self._reserved_keys,
                    ddefparams_data=self._ddef['params'].get('data'),
                    data_none=self._data_none,
                    max_ndim=self._max_ndim,
                )

    # ---------------------
    # Methods for getting a subset of the collection
    # ---------------------

    # TBC
    def get_drefddata_as_input(self, key=None, ind=None, group=None):
        lk = self._ind_tofrom_key(ind=ind, key=key, group=group, returnas=str)
        lkr = [kr for kr in self._dref['lkey']
               if any([kr in self._ddata['dict'][kk]['refs'] for kk in lk])]
        dref = {kr: {'data': self._ddata['dict'][kr]['data'],
                     'group': self._dref['dict'][kr]['group']} for kr in lkr}
        lkr = dref.keys()
        ddata = {kk: self._ddata['dict'][kk] for kk in lk if kk not in lkr}
        return dref, ddata

    # TBC
    def get_subset(self, key=None, ind=None, group=None, Name=None):
        if key is None and ind is None:
            return self
        else:
            dref, ddata = self.get_drefddata_as_input(key=key, ind=ind,
                                                      group=group)
            if Name is None and self.Id.Name is not None:
                Name = self.Id.Name + '-subset'
            return self.__class__(dref=dref, ddata=ddata, Name=Name)

    # ---------------------
    # Methods for exporting plot collection (subset)
    # ---------------------

    # TBC
    def to_PlotCollection(self, key=None, ind=None, group=None, Name=None,
                          dnmax=None, lib='mpl'):
        dref, ddata = self.get_drefddata_as_input(
            key=key, ind=ind, group=group,
        )
        if Name is None and self.Id.Name is not None:
            Name = self.Id.Name + '-plot'
        import tofu.data._core_plot as _core_plot
        if lib == 'mpl':
            cls = _core_plot.DataCollectionPlot_mpl
        else:
            raise NotImplementedError
        obj = cls(dref=dref, ddata=ddata, Name=Name)
        if dnmax is not None:
            obj.set_dnmax(dnmax)
        return obj

    # ---------------------
    # Methods for showing data
    # ---------------------

    def get_summary(
        self,
        show=None,
        show_core=None,
        sep='  ',
        line='-',
        just='l',
        table_sep=None,
        verb=True,
        return_=False,
    ):
        """ Summary description of the object content """
        # # Make sure the data is accessible
        # msg = "The data is not accessible because self.strip(2) was used !"
        # assert self._dstrip['strip']<2, msg

        lcol, lar = [], []

        # -----------------------
        # Build for groups
        if len(self._dgroup) > 0:
            lcol.append(['group', 'nb. ref', 'nb. data'])
            lar.append([
                (
                    k0,
                    len(self._dgroup[k0]['lref']),
                    len(self._dgroup[k0]['ldata']),
                )
                for k0 in self._dgroup.keys()
            ])

        # -----------------------
        # Build for refs
        if len(self._dref) > 0:
            lcol.append(['ref key', 'group', 'size', 'nb. data'])
            lar.append([
                (
                    k0,
                    self._dref[k0]['group'],
                    str(self._dref[k0]['size']),
                    len(self._dref[k0]['ldata'])
                )
                for k0 in self._dref.keys()
            ])

        # -----------------------
        # Build for ddata
        if len(self._ddata) > 0:
            if show_core is None:
                show_core = self._show_in_summary_core
            if isinstance(show_core, str):
                show_core = [show_core]
            lp = self.get_lparam(which='data')
            lkcore = ['shape', 'group', 'ref']
            assert all([ss in lp + lkcore for ss in show_core])
            col2 = ['data key'] + show_core

            if show is None:
                show = self._show_in_summary
            if show == 'all':
                col2 += [pp for pp in lp if pp not in col2]
            else:
                if isinstance(show, str):
                    show = [show]
                assert all([ss in lp for ss in show])
                col2 += [pp for pp in show if pp not in col2]

            ar2 = []
            for k0 in self._ddata.keys():
                lu = [k0] + [str(self._ddata[k0].get(cc)) for cc in col2[1:]]
                ar2.append(lu)

            lcol.append(col2)
            lar.append(ar2)

        # -----------------------
        # Build for dref_static
        if len(self._dref_static) > 0:
            for k0, v0 in self._dref_static.items():
                lk = list(list(v0.values())[0].keys())
                col = [k0] + [pp for pp in lk]
                ar = [
                    tuple([k1] + [str(v1[kk]) for kk in lk])
                    for k1, v1 in v0.items()
                ]
                lcol.append(col)
                lar.append(ar)

        # -----------------------
        # Build for dobj
        if len(self._dobj) > 0:
            for k0, v0 in self._dobj.items():
                lk = self.get_lparam(which=k0)
                lk = [
                    kk for kk in lk
                    if 'func' not in kk
                    and 'class' not in kk
                ]
                lcol.append([k0] + [pp for pp in lk])
                lar.append([
                    tuple([k1] + [str(v1[kk]) for kk in lk])
                    for k1, v1 in v0.items()
                ])

        return self._get_summary(
            lar,
            lcol,
            sep=sep,
            line=line,
            table_sep=table_sep,
            verb=verb,
            return_=return_,
        )

    # -----------------
    # conversion wavelength - energy - frequency
    # ------------------

    @staticmethod
    def convert_spectral(
        data=None,
        units_in=None, units_out=None,
        returnas=None,
    ):
        """ convert wavelength / energy/ frequency

        Available units:
            wavelength: m, mm, um, nm, A
            energy:     J, eV, keV
            frequency:  Hz, kHz, MHz, GHz

        Can also just return the conversion coef if returnas='coef'
        """
        return _comp_spectrallines.convert_spectral(
            data_in=data, units_in=units_in, units_out=units_out,
            returnas=returnas,
        )

    # -----------------
    # Get common ref
    # ------------------

    def _get_common_ref_data_nearest(
        self,
        group=None,
        lkey=None,
        return_all=None,
    ):
        """ Typically used to get a common (intersection) time vector

        Returns a time vector that contains all time points from all data
        Also return a dict of indices to easily remap each time vector to tall
            such that t[ind] = tall (with nearest approximation)

        """
        return _DataCollection_comp._get_unique_ref_dind(
            dd=self._ddata, group=group,
            lkey=lkey, return_all=return_all,
        )

    def _get_pts_from_mesh(self, key=None):
        """ Get default pts from a mesh """

        # Check key is relevant
        c0 = (
            key in self._ddata.keys()
            and isinstance(self._ddata[key].get('data'), dict)
            and 'type' in self._ddata[key]['data'].keys()
        )
        if not c0:
            msg = (
                "ddata['{}'] does not exist or is not a mesh".format(key)
            )
            raise Exception(msg)

        if self.ddata[key]['data']['type'] == 'rect':
            if self.ddata[key]['data']['shapeRZ'] == ('R', 'Z'):
                R = np.repeat(self.ddata[key]['data']['R'],
                              self.ddata[key]['data']['nZ'])
                Z = np.tile(self.ddata[key]['data']['Z'],
                            self.ddata[key]['data']['nR'])
            else:
                R = np.tile(self.ddata[key]['data']['R'],
                            self.ddata[key]['data']['nZ'])
                Z = np.repeat(self.ddata[key]['data']['Z'],
                              self.ddata[key]['data']['nR'])
            pts = np.array([R, np.zeros((R.size,)), Z])
        else:
            pts = self.ddata[key]['data']['nodes']
            pts = np.array([
                pts[:, 0], np.zeros((pts.shape[0],)), pts[:, 1],
            ])
        return pts

    # ---------------------
    # Method for interpolation - inputs checks
    # ---------------------

    # Useful?
    @property
    def _get_lquant_both(self, group1d=None, group2d=None):
        """ Return list of quantities available both in 1d and 2d """
        lq1 = [
            self._ddata[vd]['quant'] for vd in self._dgroup[group1d]['ldata']
        ]
        lq2 = [
            self._ddata[vd]['quant'] for vd in self._dgroup[group2d]['ldata']
        ]
        lq = list(set(lq1).intersection(lq2))
        return lq

    def _check_qr12RPZ(
        self,
        quant=None,
        ref1d=None,
        ref2d=None,
        q2dR=None,
        q2dPhi=None,
        q2dZ=None,
        group1d=None,
        group2d=None,
    ):

        if group1d is None:
            group1d = self._group1d
        if group2d is None:
            group2d = self._group2d

        lc0 = [quant is None, ref1d is None, ref2d is None]
        lc1 = [q2dR is None, q2dPhi is None, q2dZ is None]
        if np.sum([all(lc0), all(lc1)]) != 1:
            msg = (
                "Please provide either (xor):\n"
                + "\t- a scalar field (isotropic emissivity):\n"
                + "\t\tquant : scalar quantity to interpolate\n"
                + "\t\t\tif quant is 1d, intermediate reference\n"
                + "\t\t\tfields are necessary for 2d interpolation\n"
                + "\t\tref1d : 1d reference field on which to interpolate\n"
                + "\t\tref2d : 2d reference field on which to interpolate\n"
                + "\t- a vector (R,Phi,Z) field (anisotropic emissivity):\n"
                + "\t\tq2dR :  R component of the vector field\n"
                + "\t\tq2dPhi: R component of the vector field\n"
                + "\t\tq2dZ :  Z component of the vector field\n"
                + "\t\t=> all components have the same time and mesh!\n"
            )
            raise Exception(msg)

        # Check requested quant is available in 2d or 1d
        if all(lc1):
            (
                idquant, idref1d, idref2d,
            ) = _DataCollection_check_inputs._get_possible_ref12d(
                dd=self._ddata,
                key=quant, ref1d=ref1d, ref2d=ref2d,
                group1d=group1d,
                group2d=group2d,
            )
            idq2dR, idq2dPhi, idq2dZ = None, None, None
            ani = False
        else:
            idq2dR, msg = _DataCollection_check_inputs._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dR, group=group2d, msgstr='quant', raise_=True,
            )
            idq2dPhi, msg = _DataCollection_check_inputs._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dPhi, group=group2d, msgstr='quant', raise_=True,
            )
            idq2dZ, msg = _DataCollection_check_inputs._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dZ, group=group2d, msgstr='quant', raise_=True,
            )
            idquant, idref1d, idref2d = None, None, None
            ani = True
        return idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani

    # ---------------------
    # Method for interpolation
    # ---------------------

    def _get_finterp(
        self,
        idquant=None, idref1d=None, idref2d=None, idmesh=None,
        idq2dR=None, idq2dPhi=None, idq2dZ=None,
        interp_t=None, interp_space=None,
        fill_value=None, ani=None, Type=None,
        group0d=None, group2d=None,
    ):

        if interp_t is None:
            interp_t = 'nearest'
        if interp_t != 'nearest':
            msg = "'nearest' is the only time-interpolation method available"
            raise NotImplementedError(msg)
        if group0d is None:
            group0d = self._group0d
        if group2d is None:
            group2d = self._group2d

        # Get idmesh
        if idmesh is None:
            if idquant is not None:
                # isotropic
                if idref1d is None:
                    lidmesh = [qq for qq in self._ddata[idquant]['ref']
                               if self._dref[qq]['group'] == group2d]
                else:
                    lidmesh = [qq for qq in self._ddata[idref2d]['ref']
                               if self._dref[qq]['group'] == group2d]
            else:
                # anisotropic
                assert idq2dR is not None
                lidmesh = [qq for qq in self._ddata[idq2dR]['ref']
                           if self._dref[qq]['group'] == group2d]
            assert len(lidmesh) == 1
            idmesh = lidmesh[0]

        # Get common time indices
        if interp_t == 'nearest':
            tall, tbinall, ntall, dind = _DataCollection_comp._get_tcom(
                idquant, idref1d, idref2d, idq2dR,
                dd=self._ddata, group=group0d,
            )

        # Get mesh
        if self._ddata[idmesh]['data']['type'] == 'rect':
            mpltri = None
            trifind = self._ddata[idmesh]['data']['trifind']
        else:
            mpltri = self._ddata[idmesh]['data']['mpltri']
            trifind = mpltri.get_trifinder()

        # # Prepare output

        # Interpolate
        # Note : Maybe consider using scipy.LinearNDInterpolator ?
        if idquant is not None:
            vquant = self._ddata[idquant]['data']
            c0 = (
                self._ddata[idmesh]['data']['type'] == 'quadtri'
                and self._ddata[idmesh]['data']['ntri'] > 1
            )
            if c0:
                vquant = np.repeat(
                    vquant,
                    self._ddata[idmesh]['data']['ntri'],
                    axis=0,
                )
            vr1 = self._ddata[idref1d]['data'] if idref1d is not None else None
            vr2 = self._ddata[idref2d]['data'] if idref2d is not None else None

            # add time dimension if none
            if vquant.ndim == 1:
                vquant = vquant[None, :]
            if vr1.ndim == 1:
                vr1 = vr1[None, :]
            if vr2.ndim == 1:
                vr2 = vr2[None, :]

        else:
            vq2dR = self._ddata[idq2dR]['data']
            vq2dPhi = self._ddata[idq2dPhi]['data']
            vq2dZ = self._ddata[idq2dZ]['data']

            # add time dimension if none
            if vq2dR.ndim == 1:
                vq2dR = vq2dR[None, :]
            if vq2dPhi.ndim == 1:
                vq2dPhi = vq2dPhi[None, :]
            if vq2dZ.ndim == 1:
                vq2dZ = vq2dZ[None, :]

        if interp_space is None:
            interp_space = self._ddata[idmesh]['data']['ftype']

        # get interpolation function
        if ani:
            # Assuming same mesh and time vector for all 3 components
            func = _comp.get_finterp_ani(
                idq2dR, idq2dPhi, idq2dZ,
                interp_t=interp_t,
                interp_space=interp_space,
                fill_value=fill_value,
                idmesh=idmesh, vq2dR=vq2dR,
                vq2dZ=vq2dZ, vq2dPhi=vq2dPhi,
                tall=tall, tbinall=tbinall, ntall=ntall,
                indtq=dind.get(idquant),
                trifind=trifind, Type=Type, mpltri=mpltri,
            )
        else:
            func = _comp.get_finterp_isotropic(
                idquant, idref1d, idref2d,
                vquant=vquant, vr1=vr1, vr2=vr2,
                interp_t=interp_t,
                interp_space=interp_space,
                fill_value=fill_value,
                idmesh=idmesh,
                tall=tall, tbinall=tbinall, ntall=ntall,
                mpltri=mpltri, trifind=trifind,
                indtq=dind.get(idquant),
                indtr1=dind.get(idref1d), indtr2=dind.get(idref2d),
            )

        return func

    def _interp_pts2d_to_quant1d(
        self,
        pts=None,
        vect=None,
        t=None,
        quant=None,
        ref1d=None,
        ref2d=None,
        q2dR=None,
        q2dPhi=None,
        q2dZ=None,
        interp_t=None,
        interp_space=None,
        fill_value=None,
        Type=None,
        group0d=None,
        group1d=None,
        group2d=None,
        return_all=None,
    ):
        """ Return the value of the desired 1d quantity at 2d points

        For the desired inputs points (pts):
            - pts are in (X, Y, Z) coordinates
            - space interpolation is linear on the 1d profiles
        At the desired input times (t):
            - using a nearest-neighbourg approach for time

        """
        # Check inputs
        if group0d is None:
            group0d = self._group0d
        if group1d is None:
            group1d = self._group1d
        if group2d is None:
            group2d = self._group2d
        # msg = "Only 'nearest' available so far for interp_t!"
        # assert interp_t == 'nearest', msg

        # Check requested quant is available in 2d or 1d
        idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani = \
                self._check_qr12RPZ(
                    quant=quant, ref1d=ref1d, ref2d=ref2d,
                    q2dR=q2dR, q2dPhi=q2dPhi, q2dZ=q2dZ,
                    group1d=group1d, group2d=group2d,
                )

        # Check the pts is (3,...) array of floats
        idmesh = None
        if pts is None:
            # Identify mesh to get default points
            if ani:
                idmesh = [id_ for id_ in self._ddata[idq2dR]['ref']
                          if self._dref[id_]['group'] == group2d][0]
            else:
                if idref1d is None:
                    idmesh = [id_ for id_ in self._ddata[idquant]['ref']
                              if self._dref[id_]['group'] == group2d][0]
                else:
                    idmesh = [id_ for id_ in self._ddata[idref2d]['ref']
                              if self._dref[id_]['group'] == group2d][0]

            # Derive pts
            pts = self._get_pts_from_mesh(key=idmesh)

        pts = np.atleast_2d(pts)
        if pts.shape[0] != 3:
            msg = (
                "pts must be np.ndarray of (X,Y,Z) points coordinates\n"
                + "Can be multi-dimensional, but 1st dimension is (X,Y,Z)\n"
                + "    - Expected shape : (3,...)\n"
                + "    - Provided shape : {}".format(pts.shape)
            )
            raise Exception(msg)

        # Check t
        lc = [t is None, type(t) is str, type(t) is np.ndarray]
        assert any(lc)
        if lc[1]:
            assert t in self._ddata.keys()
            t = self._ddata[t]['data']

        # Interpolation (including time broadcasting)
        # this is the second slowest step (~0.08 s)
        func = self._get_finterp(
            idquant=idquant, idref1d=idref1d, idref2d=idref2d,
            idq2dR=idq2dR, idq2dPhi=idq2dPhi, idq2dZ=idq2dZ,
            idmesh=idmesh,
            interp_t=interp_t, interp_space=interp_space,
            fill_value=fill_value, ani=ani, Type=Type,
            group0d=group0d, group2d=group2d,
        )

        # Check vect of ani
        c0 = (
            ani is True
            and (
                vect is None
                or not (
                    isinstance(vect, np.ndarray)
                    and vect.shape == pts.shape
                )
            )
        )
        if c0:
            msg = (
                "Anisotropic field interpolation needs a field of local vect\n"
                + "  => Please provide vect as (3, npts) np.ndarray!"
            )
            raise Exception(msg)

        # This is the slowest step (~1.8 s)
        val, t = func(pts, vect=vect, t=t)

        # return
        if return_all is None:
            return_all = True
        if return_all is True:
            dout = {
                't': t,
                'pts': pts,
                'ref1d': idref1d,
                'ref2d': idref2d,
                'q2dR': idq2dR,
                'q2dPhi': idq2dPhi,
                'q2dZ': idq2dZ,
                'interp_t': interp_t,
                'interp_space': interp_space,
            }
            return val, dout
        else:
            return val

    # TBC
    def _interp_one_dim(x=None, ind=None, key=None, group=None,
                        kind=None, bounds_error=None, fill_value=None):
        """ Return a dict of interpolated data

        Uses scipy.inpterp1d with args:
            - kind, bounds_error, fill_value

        The interpolated data is chosen method select() with args:
            - key, ind

        The interpolation is done against a reference vector x
            - x can be a key to an existing ref
            - x can be user-provided array
                in thay case the group should be specified
                (to properly identify the interpolation dimension)

        Returns:
        --------
        dout:       dict
            dict of interpolated data
        dfail:  dict of failed interpolations, with error messages

        """

        # Check x
        assert x is not None
        if isinstance(x) is str:
            if x not in self.lref:
                msg = "If x is a str, it must be a valid ref!\n"
                msg += "    - x: {}\n".format(x)
                msg += "    - self.lref: {}".format(self.lref)
                raise Exception(msg)
            group = self._dref[x]['group']
            x = self._ddata[x]['data']
        else:
            try:
                x = np.atleast_1d(x).ravel()
            except Exception:
                msg = (
                    "The reference with which to interpolate, x, should be:\n"
                    + "    - a key to an existing ref\n"
                    + "    - a 1d np.ndarray"
                )
                raise Exception(x)
            if group not in self.lgroup:
                msg = "Interpolation must be with respect to a group\n"
                msg += "Provided group is not in self.lgroup:\n"
                msg += "    - group: {}".format(group)
                raise Exception(msg)

        # Get keys to interpolate
        if ind is None and key in None:
            lk = self._dgroup[group]['ldata']
        else:
            lk = self._ind_tofrom_key(ind=ind, key=key, returnas=str)

        # Check provided keys are relevant, and get dim index
        dind, dfail = {}, {}
        for kk in lk:
            if kk not in self._dgroup[group]['ldata']:
                # gps = self._ddata[kk]['groups']
                # msg = "Some data not in interpolation group:\n"
                # msg += "    - self.ddata[%s]['groups'] = %s"%(kk,str(gps))
                # msg += "    - Interpolation group: %s"%group
                # raise Exception(msg)
                dfail[kk] = "Not dependent on group {}".format(group)
            else:
                dind[kk] = self._ddata[kk]['groups'].index(group)

        # Start loop for interpolation
        dout = {}
        for kk in dout.keys():
            shape = self._ddata['dict'][kk]['shape']

            if not isinstance(self._ddata[kk]['data'], np.ndarray):
                dfail[kk] = "Not a np.ndarray !"
                continue

            kr = self._ddata['dict'][kk]['refs'][dind[kk]]
            vr = self._ddata['dict'][kr]['data']
            data = self._ddata['dict'][kk]['data']
            try:
                if dind[kk] == len(shape) - 1:
                    dout[kk] = scpinterp.interp1d(vr, y,
                                                  kind=kind, axis=-1,
                                                  bounds_error=bounds_error,
                                                  fill_value=fill_value,
                                                  assume_sorted=True)(x)
                else:
                    dout[kk] = scpinterp.interp1d(vr, y,
                                                  kind=kind, axis=dind[kk],
                                                  bounds_error=bounds_error,
                                                  fill_value=fill_value,
                                                  assume_sorted=True)(x)

            except Exception as err:
                dfail[kk] = str(err)
        return dout, dfail

    # ---------------------
    # Method for fitting models in one direction
    # ---------------------

    # TBC
    def _fit_one_dim(ind=None, key=None, group=None,
                     Type=None, func=None, **kwdargs):
        """ Return the parameters of a fitted function

        The interpolated data is chosen method select() with args:
            - key, ind

        Returns:
        --------
        dout:       dict
            dict of interpolated data
        dfail:  dict of failed interpolations, with error messages

        """
        # Get keys to interpolate
        lk = self._ind_tofrom_key(ind=ind, key=key, group=group, returnas=str)

        # Start model fitting loop on data keys
        dout = {}
        for kk in lk:
            x = None
            axis = None
            dfit = _DataCollection_comp.fit(
                self._ddata['dict'][kk]['data'],
                x=x,
                axis=axis,
                func=func,
                Type=Type,
                **kwdargs,
            )
            dout[kk] = dfit

        return dout

    # ---------------------
    # Methods for plotting data
    # ---------------------

    def plot_as_matrix(
        self,
        key=None,
        ind=None,
        vmin=None,
        vmax=None,
        cmap=None,
        aspect=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
    ):
        """ Plot the desired 2d data array as a matrix """
        return _DataCollection_plot.plot_as_matrix(
            coll=self,
            key=key,
            ind=ind,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            aspect=aspect,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
        )

    def _plot_timetraces(self, ntmax=1, group='time',
                         key=None, ind=None, Name=None,
                         color=None, ls=None, marker=None, ax=None,
                         axgrid=None, fs=None, dmargin=None,
                         legend=None, draw=None, connect=None, lib=None):
        plotcoll = self.to_PlotCollection(ind=ind, key=key, group=group,
                                          Name=Name, dnmax={group: ntmax})
        return _DataCollection_plot.plot_DataColl(
            plotcoll,
            color=color, ls=ls, marker=marker, ax=ax,
            axgrid=axgrid, fs=fs, dmargin=dmargin,
            draw=draw, legend=legend,
            connect=connect, lib=lib,
        )

    def _plot_axvlines(
        self,
        which=None,
        key=None,
        ind=None,
        param_x=None,
        param_txt=None,
        sortby=None,
        sortby_def=None,
        sortby_lok=None,
        ax=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        dsize=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):
        """ plot rest wavelengths as vertical lines """

        # Check inputs
        which, dd = self.__check_which(
            which=which, return_dict=True,
        )
        key = self._ind_tofrom_key(which=which, key=key, ind=ind, returnas=str)

        if sortby is None:
            sortby = sortby_def
        if sortby not in sortby_lok:
            msg = (
                """
                For plotting, sorting can be done only by:
                {}

                You provided:
                {}
                """.format(sortby_lok, sortby)
            )
            raise Exception(msg)

        return _DataCollection_plot.plot_axvline(
            din=dd,
            key=key,
            param_x='lambda0',
            param_txt='symbol',
            sortby=sortby, dsize=dsize,
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor,
            fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

    # ---------------------
    # saving => get rid of function
    # ---------------------

    def save(self, path=None, name=None,
             strip=None, sep=None, deep=True, mode='npz',
             compressed=False, verb=True, return_pfe=False):

        # Remove function mpltri if relevant
        lk = [
            k0 for k0, v0 in self._ddata.items()
            if isinstance(v0['data'], dict)
            and 'mpltri' in v0['data'].keys()
        ]
        for k0 in lk:
            del self._ddata[k0]['data']['mpltri']
        lk = [
            k0 for k0, v0 in self._ddata.items()
            if isinstance(v0['data'], dict)
            and 'trifind' in v0['data'].keys()
        ]
        for k0 in lk:
            del self._ddata[k0]['data']['trifind']

        # call parent method
        return super().save(
            path=path, name=name,
            sep=sep, deep=deep, mode=mode,
            strip=strip, compressed=compressed,
            return_pfe=return_pfe, verb=verb
        )




"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import os
import warnings


# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu.data as tfd
import tofu.utils as tfu

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.data.DataCollection'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


# def my_setup_function():
#    print ("my_setup_function")

# def my_teardown_function():
#    print ("my_teardown_function")

# @with_setup(my_setup_function, my_teardown_function)
# def test_numbers_3_4():
#    print 'test_numbers_3_4  <============================ actual test code'
#    assert multiply(3,4) == 12

# @with_setup(my_setup_function, my_teardown_function)
# def test_strings_a_3():
#    print 'test_strings_a_3  <============================ actual test code'
#    assert multiply('a',3) == 'aaa'


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_DataCollection(object):

    @classmethod
    def setup_class(cls, Name='data1',  SavePath='./', verb=False):

        # time vectors
        t0 = np.linspace(0, 10, 10)
        t1 = np.linspace(0, 10, 50)
        t2 = np.linspace(0, 10, 100)
        cls.lt = [t0, t1, t2]

        # radii vectors
        r0 = np.linspace(0, 1, 10)
        r1 = np.linspace(0, 1, 50)
        r2 = np.linspace(0, 1, 200)
        cls.lr = [r0, r1, r2]

        # chan
        ch0 = np.arange(0, 2)
        ch1 = np.arange(0, 5)
        cls.lch = [ch0, ch1]

        # meshes
        mesh0 = {
            'type': 'rect',
            'R': np.r_[0, 1, 2, 3],
            'Z': np.r_[0, 1, 2],
            'shapeRZ': ('R', 'Z'),
        }
        mesh1 = {
            'type': 'tri',
            'nodes': np.array([[0, 1, 1, 0], [0, 0, 1, 1]]).T,
            'faces': np.array([[0, 1, 2], [2, 3, 0]]),
        }
        cls.lmesh = [mesh0, mesh1]

        # traces
        trace00, trace01 = 2.*t0, np.sin(t0)
        trace10, trace11 = np.cos(t1), t1[:, None]*t0
        trace20, trace21 = np.sin(r0), r0[:, None]*r1
        trace30, trace31 = np.cos(r2), t0[:, None]*np.sin(r2)
        trace40 = t2[:, None, None]*r1[None, :, None]*ch0[None, None, :]
        trace41 = t2[None, None, :]*r2[:, None, None]*ch1[None, :, None]
        trace50 = np.cos(t0)[:, None, None]*(
            mesh0['R'][None, :, None]*mesh0['Z'][None, None, :]
        )
        trace51 = t1[:, None]*(mesh1['faces'][:, 0:1]).T
        cls.ltrace = [trace00, trace01, trace10, trace11,
                      trace20, trace21, trace30, trace31,
                      trace40, trace41, trace50, trace51]

        # Anisotropic rect + tri
        BRr = np.cos(t1)[:, None, None]*(
            mesh0['R'][None, :, None]*mesh0['Z'][None, None, :]
        )
        BPhir = np.cos(t1)[:, None, None]*(
            mesh0['R'][None, :, None]*mesh0['Z'][None, None, :]
        )
        BZr = np.cos(t1)[:, None, None]*(
            mesh0['R'][None, :, None]*mesh0['Z'][None, None, :]
        )
        BRt = t0[:, None]*(mesh1['faces'][:, 0:1]).T
        BPhit = t0[:, None]*(mesh1['faces'][:, 0:1]).T
        BZt = t0[:, None]*(mesh1['faces'][:, 0:1]).T
        cls.lB = [BRr, BPhir, BZr, BRt, BPhit, BZt]

        # polygons
        lpoly0 = [np.ones((2, 5)), np.ones((2, 8))]
        lpoly1 = [np.ones((2, 5)), np.ones((2, 8)), np.ones((2, 5))]
        lpoly2 = [np.ones((2, 5)), np.ones((2, 5))]
        cls.lpoly = [lpoly0, lpoly1, lpoly2]

        # spectral lines
        l0 = {
            'key': 'l0', 'lambda0': 5e-10,
            'origin': '[1]', 'transition': 'A->B',
        }
        l1 = {
            'key': 'l1', 'lambda0': 5e-10,
            'origin': '[2]', 'transition': 'B->C',
        }
        l2 = {
            'key': 'l2',
            'data': t0[:, None]*t1[None, :], 'ref': ('t0', 't1'),
            'lambda0': 5e-10, 'origin': '[2]', 'transition': 'B->C'
        }
        cls.llines = [l0, l1, l2]

        # Configs
        # conf0 = tfg.utils.create_config(case='B2')
        # conf1 = tfg.utils.create_config(case='B3')

        dref = {
            't0': {
                'data': cls.lt[0], 'group': 'time', 'units': 's',
                'quant': 'time',
            },
            't1': {
                'data': cls.lt[1], 'group': 'time', 'units': 'min',
                'quant': 'time',
            },
            'r2': {
                'data': cls.lr[2],
                'group': 'radius', 'units': 'm', 'quant': 'rho',
            },
            'mesh0': {
                'data': mesh0,
            },
            'mesh1': {
                'data': mesh1,
            },
        }
        ddata = {
            'trace00': {'data': cls.ltrace[0], 'ref': ('t0',)},
            'trace10': {'data': cls.ltrace[2], 'ref': ('t1',), 'units': 's'},
            'trace11': {'data': cls.ltrace[3], 'ref': ('t1', 't0')},
            'trace30': {'data': cls.ltrace[6], 'ref': ('r2',)},
            'trace31': {'data': cls.ltrace[7], 'ref': ('t0', 'r2')},
            'trace50': {'data': trace50, 'ref': ('t0', 'mesh0')},
            'BRr': {'data': BRr, 'ref': ('t1', 'mesh0'), 'quant': 'BR'},
            'BPhir': {'data': BPhir, 'ref': ('t1', 'mesh0'), 'quant': 'BPhi'},
            'BZr': {'data': BZr, 'ref': ('t1', 'mesh0'), 'quant': 'BZ'},
            'BRt': {'data': BRt, 'ref': ('t0', 'mesh1'), 'quant': 'BR'},
            'BPhit': {'data': BPhit, 'ref': ('t0', 'mesh1'), 'quant': 'BPhi'},
            'BZt': {'data': BZt, 'ref': ('t0', 'mesh1'), 'quant': 'BZ'},
        }
        data = tfd.DataCollection(dref=dref, ddata=ddata, Name=Name)

        # Spectrallines
        dref = {
            't0': {'data': cls.lt[0], 'group': 'time', 'units': 's'},
            't1': {'data': cls.lt[1], 'group': 'time', 'units': 'min'},
        }
        dref_static = {
            'source': {
                '[1]': {'long': 'blabla'},
                '[2]': {'long': 'blibli'},
            },
            'ion': {
                'O3+': {'element': 'O'},
                'Ca6+': {'element': 'Ca'},
            },
        }
        dobj = {
            'lines': {
                'l0': {
                    'lambda0': 3e-10, 'source': '[1]', 'transition': 'A->B',
                },
                'l1': {
                    'lambda0': 4e-10, 'source': '[2]', 'transition': 'B->C',
                },
                'l2': {
                    'data': t0[:, None]*t1[None, :], 'ref': ('t0', 't1'),
                    'lambda0': 5e-10, 'source': '[2]', 'transition': 'B->C',
                },
            }
        }
        sl = tfd.DataCollection()
        sl._data_none = True
        sl.update(dref=dref, dref_static=dref_static, dobj=dobj)

        cls.lobj = [data, sl]

    @classmethod
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_init_from_combinations(self):

        # Try with minimalist input (implicit with n = 1)
        dgroup = 'time'
        dref = {'t0': self.lt[0]}
        ddata = {
            'trace00': self.ltrace[0],
            'trace01': {'data': self.ltrace[1], 'units': 'a.u.'},
        }
        data = tfd.DataCollection(
            dgroup=dgroup, dref=dref, ddata=ddata,
            Name='data',
        )

        # Try with minimalist input
        dref = {'t0': {'data': self.lt[0], 'group': 'time'},
                't1': {'data': self.lt[1], 'group': 'time', 'units': 's'},
                'r2': {'data': self.lr[2], 'group': 'radius', 'foo': 'bar'}}
        ddata = {
            'trace00': {'data': self.ltrace[0], 'ref': 't0'},
            'trace10': {'data': self.ltrace[2], 'ref': 't1', 'units': 'a'},
            'trace11': {'data': self.ltrace[3], 'ref': ('t1', 't0')},
            'trace30': {'data': self.ltrace[6], 'ref': ('r2',), 'foo': 'bar'},
            'trace31': {'data': self.ltrace[7], 'ref': ('t0', 'r2')}
        }
        data = tfd.DataCollection(
            dgroup=None, dref=dref, ddata=ddata,
            Name='data',
        )

        # Try with meshes
        dref = {
            't0': {'data': self.lt[0], 'group': 'time', 'units': 's'},
            't1': {'data': self.lt[1], 'group': 'time', 'units': 's'},
            'r2': {'data': self.lr[2], 'group': 'radius', 'foo': 'bar'},
            'mesh1': {'data': self.lmesh[1], 'foo': 'bar', 'quant': 'rho'},
        }
        ddata = {
            'trace10': {'data': self.ltrace[2], 'ref': 't1', 'units': 'a'},
            'trace50': {'data': self.ltrace[-2], 'ref': ('t0', 'mesh0')},
            'trace51': {'data': self.ltrace[-1], 'ref': ('t1', 'mesh1')},
            'mesh0': {'data': self.lmesh[0], 'foo': 'bar', 'group': 'mesh2d'},
        }
        data = tfd.DataCollection(
            dref=dref, ddata=ddata,
            Name='data',
        )

        # Try with lines
        data = tfd.DataCollection()
        data.add_data(**self.llines[0])
        data.add_data(**self.llines[1])
        data.add_ref(key='t0', data=self.lt[0], group='ne')
        data.add_ref(key='t1', data=self.lt[1], group='Te')
        data.add_data(**self.llines[2])

    def test02_wrong_init(self):
        # Try with minimalist input
        dref = {
            't0': {'data': self.lt[0], 'group': 'time'},
            't1': {'data': self.lt[1], 'group': 'time'},
        }
        ddata = {
            'trace00': self.ltrace[0], 'ref': 't0',
            'trace11': self.ltrace[3], 'ref': ('t0', 't1'),
        }
        err = False
        try:
            data = tfd.DataCollection(
                dgroup=None, dref=dref, ddata=ddata,
                Name='data',
            )
        except Exception as er:
            err = True
        assert err, "Exception was not detected properly!"

    def test03_add_remove_refdataobj(self):
        data = self.lobj[0]

        data.add_ref(key='r0', data=self.lr[0], group='radius', foo='bar')
        assert 'r0' in data.dref.keys()

        data.remove_ref(key='t0')
        assert 't0' not in data.dref.keys()
        assert 't0' not in data.ddata.keys()
        assert all([tt not in data.ddata.keys()
                    for tt in ['trace00', 'trace11', 'trace31']])

        data.add_ref('t0', data=self.lt[0], group='time')
        assert 't0' in data.dref.keys()

        # Check ambiguous throws error
        err = False
        try:
            data.add_data(key='trace00', data=self.ltrace[0])
        except Exception:
            err = True
        assert err
        data.add_data('trace00', data=self.ltrace[0], ref=('t0',))
        data.add_data('trace11', data=self.ltrace[3], ref=('t1', 't0'))
        data.add_data('trace31', data=self.ltrace[7], ref=('t0', 'r2'),
                      foo='bar')
        assert all([tt in data.ddata.keys()
                    for tt in ['trace00', 'trace11', 'trace31']])

        # Add/remove mesh
        data.add_ref(key='mesh0', data=self.lmesh[0], group='mesh2d')
        data.add_data(
            key='trace51', data=self.ltrace[-1],
            ref=('t1', 'mesh1'), quant='rho',
        )
        data.add_data(
            key='BRt', data=self.lB[3], ref=('t0', 'mesh1'),
            quant='BR', units='T',
        )
        data.add_data(
            key='BPhit', data=self.lB[4], ref=('t0', 'mesh1'),
            quant='BPhi', units='T',
        )
        data.add_data(
            key='BZt', data=self.lB[5], ref=('t0', 'mesh1'),
            quant='BZ', units='T',
        )

        # Add / remove obj and ref_static
        self.lobj[1].add_ref_static(key='[3]', which='source', long='bloblo')
        self.lobj[1].add_obj(
            which='lines', key='l3',
            lambda0=5e-10, source='[3]', transition='C->D',
        )
        self.lobj[1].remove_obj(key='l3')
        self.lobj[1].remove_ref_static(key='[3]')
        self.lobj[1].remove_ref_static(which='ion')

    def test04_select(self):
        data = self.lobj[0]

        key = data.select(which='data', units='s', returnas=str)
        assert key == ['trace10']

        out = data.select(units='a.u.', returnas=int)
        assert len(out) == 12, out

        # test quantitative param selection
        out = self.lobj[1].select(which='lines', lambda0=[3.5e-10, 6e-10])
        assert len(out) == 2

        out = self.lobj[1].select(which='lines', lambda0=(3.5e-10, 6e-10))
        assert len(out) == 1

    def test05_sortby(self):
        for oo in self.lobj:
            oo.sortby(which='data', param='units')

    def test06_get_summary(self):
        for oo in self.lobj:
            oo.get_summary()

    def test07_getsetaddremove_param(self):
        data = self.lobj[0]

        out = data.get_param('units')
        data.set_param('units', value='T', key='trace00')
        data.add_param('shot', value=np.arange(0, len(data.ddata)))
        assert np.all(
            data.get_param('shot')['shot'] == np.arange(0, len(data.ddata))
        )
        data.remove_param('shot')
        assert 'shot' not in data.get_lparam(which='data')

    def test08_switch_ref(self):

        data = self.lobj[0]
        data.switch_ref('trace00')

        # Check t0 removed
        assert 'trace00' in data.dref.keys()
        assert 'trace00' in data.dgroup['time']['lref']
        assert all(['trace00' in v0['ref'] for k0, v0 in data.ddata.items()
                    if k0 in data.dref['trace00']['ldata']])
        # Check t0 removed
        assert 't0' not in data.dref.keys()
        assert 't0' not in data.dgroup['time']['lref']
        assert all(['t0' not in v0['ref'] for k0, v0 in data.ddata.items()
                    if k0 in data.dref['trace00']['ldata']])
        # .. but still in data
        assert 't0' in data.ddata.keys()

    def test09_convert_spectral(self):
        coef, inv = self.lobj[0].convert_spectral(
            units_in='eV', units_out='J', returnas='coef',
        )
        assert coef == scpct.e and inv is False

        coef, inv = self.lobj[0].convert_spectral(
            units_in='nm', units_out='keV', returnas='coef',
        )
        assert coef == (0.001*(1/scpct.e)*scpct.h*scpct.c / 1.e-9)
        assert inv is True

        data = [[0], [1], [2], [3]]
        out = self.lobj[0].convert_spectral(
            data=data,
            units_in='A', units_out='MHz',
        )
        assert out.shape == (4, 1)

    # ------------------------
    #   Interpolation tools
    # ------------------------

    def test10_check_qr12RPZ(self):
        data = self.lobj[0]
        # Directly get 2d quant
        out = data._check_qr12RPZ(
             quant='mesh0', ref1d=None, ref2d=None,
             q2dR=None, q2dPhi=None, q2dZ=None,
         )
        assert (
            out[0] == 'mesh0'
            and all([out[ii] is None for ii in [1, 2, 3, 4, 5]])
        )
        out = data._check_qr12RPZ(
             quant='trace51', ref1d=None, ref2d=None,
             q2dR=None, q2dPhi=None, q2dZ=None,
         )
        assert (
            out[0] == 'trace51'
            and all([out[ii] is None for ii in [1, 2, 3, 4, 5]])
        )

        # Get 1d quant
        out = data._check_qr12RPZ(
             quant='r2', ref1d='r2', ref2d='trace51',
             q2dR=None, q2dPhi=None, q2dZ=None,
         )
        assert (
            out[0] == 'r2' and out[1] == 'r2' and out[2] == 'trace51'
            and all([out[ii] is None for ii in [3, 4, 5]])
        )

        # Get 1d quant from 2d
        out = data._check_qr12RPZ(
             quant='trace30', ref1d='r2', ref2d='trace51',
             q2dR=None, q2dPhi=None, q2dZ=None,
         )
        assert (
            out[0] == 'trace30' and out[1] == 'r2' and out[2] == 'trace51'
            and all([out[ii] is None for ii in [3, 4, 5]])
        )

        # Get 1d quant from 2d
        out = data._check_qr12RPZ(
             quant='trace30', ref1d=None, ref2d=None,
             q2dR=None, q2dPhi=None, q2dZ=None,
         )
        assert (
            out[0] == 'trace30' and out[1] == 'r2' and out[2] == 'trace51'
            and all([out[ii] is None for ii in [3, 4, 5]])
        )

    def test11_interp_pts2d_to_quant1d(self):
        data = self.lobj[0]

        # Not specifying points
        val, dout = data._interp_pts2d_to_quant1d(quant='trace30')

        # Specifying wrong ref2d
        error = False
        try:
            val, dout = data._interp_pts2d_to_quant1d(
                quant='trace30', ref2d='mesh1',
            )
        except Exception as err:
            error = err
        assert isinstance(error, Exception) and 'Non-valid' in str(error)

        # Specifying points and time
        pts = np.copy(dout['pts'])*0.5
        val, dout = data._interp_pts2d_to_quant1d(
            quant='trace30', pts=pts, t=np.linspace(1, 5, 10),
        )

        # Not specifying points
        val, t = data._interp_pts2d_to_quant1d(quant='trace30')

        # anisotropic rect with time
        pts = data._get_pts_from_mesh(key='mesh0')
        vect = pts
        val, t = data._interp_pts2d_to_quant1d(
            q2dR='BRr', q2dZ='BZr', q2dPhi='BPhir',
            pts=pts, vect=vect, t=np.linspace(2, 5, 10),
        )

        # anisotropic tri
        pts = data._get_pts_from_mesh(key='mesh1')
        vect = pts
        val, t = data._interp_pts2d_to_quant1d(
            q2dR='BRt', q2dZ='BZt', q2dPhi='BPhit',
            pts=pts, vect=vect,
        )

    # ------------------------
    #   Generic TofuObject methods
    # ------------------------

    def test20_copy_equal(self):
        for oo in self.lobj:
            obj = oo.copy()
            assert obj == oo

    def test21_get_nbytes(self):
        for oo in self.lobj:
            nb, dnb = oo.get_nbytes()

    def test22_strip_nbytes(self, verb=False):
        lok = self.lobj[0].__class__._dstrip['allowed']
        nb = np.full((len(lok),), np.nan)
        for oo in self.lobj:
            for ii in lok:
                oo.strip(ii, verb=verb)
                nb[ii] = oo.get_nbytes()[0]
            assert np.all(np.diff(nb) <= 0.), nb
            for ii in lok[::-1]:
                oo.strip(ii, verb=verb)

    def test23_saveload(self, verb=False):
        for oo in self.lobj:
            if oo.Id.Name is None:
                try:
                    pfe = oo.save(deep=False, verb=verb, return_pfe=True)
                except Exception as err:
                    pass
            else:
                pfe = oo.save(deep=False, verb=verb, return_pfe=True)
                obj = tfu.load(pfe, verb=verb)
                # Just to check the loaded version works fine
                assert oo == obj
                os.remove(pfe)


# #############################################################################
# #############################################################################
#           Specific to SpectralLines
# #############################################################################


class Test02_SpectralLines(object):

    @classmethod
    def setup_class(cls, Name='data1',  SavePath='./', verb=False):
        cls.sl = tfd.SpectralLines.from_openadas(
            lambmin=3.94e-10,
            lambmax=4e-10,
            element=['Ar', 'W'],
        )

    @classmethod
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_add_from_openadas(self):
        lines = self.sl.dobj['lines']
        self.sl.add_from_openadas(
            lambmin=3.90e-10,
            lambmax=3.96e-10,
            element='W',
        )
        assert all([k0 in self.sl.dobj['lines'].keys() for k0 in lines.keys()])

    def test02_sortby(self):
        self.sl.sortby(param='lambda0', which='lines')
        self.sl.sortby(param='ion', which='lines')

    def test03_convert_lines(self):
        self.sl.convert_lines(units='Hz')

    def test04_calc_pec(self):
        ne = np.r_[1e15, 1e18, 1e21]
        Te = np.r_[1e3, 2e3, 3e3, 4e3, 5e3]
        dpec = self.sl.calc_pec(ne=ne, Te=Te[:ne.size], grid=False)
        dpec = self.sl.calc_pec(
            key='Ar16_9_oa_pec40_cl', ne=ne, Te=Te[:ne.size], grid=False,
        )
        dpec = self.sl.calc_pec(ne=ne, Te=Te, grid=True)
        dpec = self.sl.calc_pec(
            key='Ar16_9_oa_pec40_cl', ne=ne, Te=Te[:ne.size], grid=False,
        )

    def test05_calc_intensity(self):
        ne = np.r_[1e15, 1e18, 1e21]
        Te = np.r_[1e3, 2e3, 3e3, 4e3, 5e3]

        concentration = np.r_[0.1, 0.2, 0.3]
        dint = self.sl.calc_intensity(
            ne=ne, Te=Te[:ne.size], concentration=concentration, grid=False,
        )

        key = ['Ar16_9_oa_pec40_cl']
        concentration = {k0: np.r_[0.1, 0.2, 0.3] for k0 in key}
        dint = self.sl.calc_intensity(
            key=key,
            ne=ne, Te=Te[:ne.size], concentration=concentration, grid=False,
        )

        concentration = np.random.random((ne.size, Te.size))
        dint = self.sl.calc_intensity(
            ne=ne, Te=Te, concentration=concentration, grid=True,
        )

        key = ['Ar16_9_oa_pec40_cl']
        concentration = {k0: concentration for k0 in key}
        dint = self.sl.calc_intensity(
            key=key,
            ne=ne, Te=Te, concentration=concentration, grid=True,
        )

    def test06_plot(self):
        ax = self.sl.plot()
        plt.close('all')

    def test07_plot_pec_single(self):
        Te = 1.e3
        ne = 1.e20
        ax = self.sl.plot_pec_single(Te=Te, ne=ne)

    def test08_plot_pec(self):
        Te = np.linspace(1, 7, 7)*1e3
        ne = np.logspace(15, 21, 7)
        ax = self.sl.plot_pec(Te=1e3, ne=ne)
        ax = self.sl.plot_pec(Te=Te, ne=1e19)
        ax = self.sl.plot_pec(Te=Te, ne=ne)
        plt.close('all')
